"""
This file collects all functions directly related to calculating
the four major components of ZephIR's loss function:
1. corr_loss: normalized correlation loss for image registration
2. reg_n: spring network maintaining flexible connections to nearest neighbors
3. reg_t: linear temporal smoothing across a small sequence of frames
4. reg_d: nuclei probability as detected by ZephOD network
"""
import torch
from torch.nn.functional import grid_sample

from ..utils.utils import *


def corr_loss(prediction, target):
    """Image registration loss, L_R.

    Normalized correlation loss between two lists of volumes (T, C, Z, Y, X).
    Loss is calculated over (Z, Y, X) axes and averaged over (C) axis. (T) axis
    is not reduced.

    :param prediction: child descriptors
    :param target: target descriptors
    :return: loss
    """

    vx = prediction - torch.mean(prediction, dim=[2, 3, 4, 5], keepdim=True)
    vy = target - torch.mean(target, dim=[2, 3, 4, 5], keepdim=True)
    # child descriptors can sometimes be empty and cause DivByZero error
    if torch.any(torch.std(prediction, dim=[2, 3, 4, 5]) == 0):
        sxy = torch.mul(torch.std(target, dim=[2, 3, 4, 5]),
                        torch.std(target, dim=[2, 3, 4, 5]))
    else:
        sxy = torch.mul(torch.std(prediction, dim=[2, 3, 4, 5]),
                        torch.std(target, dim=[2, 3, 4, 5]))
    cc = torch.div(torch.mean(torch.mul(vx, vy), dim=[2, 3, 4, 5]), sxy + 1e-5)
    return torch.mean(1 - cc, dim=1)


def reg_n(k, rho, rho_p, ind, d_ref, subset=None, mode='disp'):
    """Spatial regularization, L_N.

    Spring-like cost to deformations in a network of nearest-neighbor connections,
    penalizing relative motion between keypoints and their neighbors, calculated
    according to given mode:
    'disp': linear cost to all relative motion
    'norm': linear cost to change in distance between keypoints, rotation is allowed
    'ljp': Lennard-Jones cost to change in distances between keypoints,
    rotation is allowed, collapsing onto the same coordinates is heavily penalized

    :param k: spring stiffnesses
    :param rho: keypoint coordinates
    :param rho_p: restricted keypoint coordinates
    :param ind: neighbor indices
    :param d_ref: neighbor displacements in reference frame
    :param subset: indices of keypoints in restricted list
    :param mode: method for calculating loss
    :return: loss
    """

    if rho_p is not None and subset is not None:
        nn_rho = torch.cat([rho_p[:, ind[:, n], :].unsqueeze(2)
                              for n in range(ind.shape[1])], dim=2)
        d = nn_rho[:, subset, ...] - rho.unsqueeze(2).expand(-1, -1, ind.shape[1], -1)
        d_ref = d_ref[subset, ...]
        k = k[subset, ...]
    else:
        nn_rho = torch.cat([rho[:, ind[:, n], :].unsqueeze(2)
                              for n in range(ind.shape[1])], dim=2)
        d = nn_rho - rho.unsqueeze(2).expand(-1, -1, ind.shape[1], -1)

    if mode == 'norm':
        d_ref = torch.norm(d_ref, dim=-1).unsqueeze(0)
        d_diff = torch.div(torch.abs(torch.norm(d, dim=-1) - d_ref), d_ref + 1e-5)
    elif mode == 'ljp':
        d_ref = torch.norm(d_ref, dim=-1).unsqueeze(0)
        d_ljp = torch.div(d_ref, torch.norm(d + 1e-5, dim=-1))
        d_diff = torch.pow(d_ljp, 4) - 2 * torch.pow(d_ljp, 2)
    else:
        d_diff = torch.div(torch.norm(d - d_ref.unsqueeze(0), dim=-1),
                           torch.norm(d_ref + 1e-5, dim=-1).unsqueeze(0))

    reg = torch.mean(k.unsqueeze(0) * d_diff, dim=-1)
    reg = [r[torch.isfinite(r)] for r in reg]
    reg = [torch.mean(r) if len(r) > 0 else torch.tensor(0).to(r.dtype) for r in reg]
    return torch.stack(reg, dim=0)


def reg_t(k, descriptors, crop_rads, npx_to_keep):
    """Temporal smoothing, L_T.

    Extracts pixel intensities at center of descriptors and penalizes change
    in intensity from the center frame.

    :param k: multiplier to modulate relative contribution to loss, lambda_T
    :param descriptors: child descriptors (T, C, Z, Y, X)
    :param crop_rads: radius of center crop (Z, Y, X)
    :param npx_to_keep: number of pixels to keep when calculating average intensity value
    :return: loss
    """

    shape_t, shape_n, shape_c = descriptors.shape[:3]
    traces = torch.zeros(shape_t, shape_n, shape_c)

    shape_zyx = np.array([descriptors.shape[3], descriptors.shape[4], descriptors.shape[5]])
    crop_i, crop_f = shape_zyx // 2 - crop_rads, shape_zyx // 2 + crop_rads + 1
    descriptors = descriptors[..., crop_i[0]:crop_f[0], :, :]
    descriptors = descriptors[..., crop_i[1]:crop_f[1], :]
    descriptors = descriptors[..., crop_i[2]:crop_f[2]]
    descriptors = descriptors.reshape(shape_t, shape_n, shape_c, -1)

    for t in range(shape_t):
        for n in range(shape_n):
            for c in range(shape_c):
                if npx_to_keep is not None:
                    roi = torch.sort(descriptors[t, n, c])[0][-npx_to_keep:]
                else:
                    roi = torch.sort(descriptors[t, n, c])[0]
                traces[t, n, c] = torch.mean(roi)

    traces = torch.abs(traces - traces[shape_t // 2, ...].unsqueeze(0))
    return k * torch.mean(torch.sum(traces, dim=-1), dim=-1)


def reg_d(k, pred, rho, n_chunks):
    """Feature detection loss, L_D.

    Calculates feature detection probability at keypoint coordinates.

    :param k: multiplier to modulate relative contribution to loss, lambda_D
    :param pred: Zephod prediction of feature detection probability map
    :param rho: keypoint coordinates
    :param n_chunks: number of chunks to divide forward process into
    :return: loss
    """
    shape_t, shape_n = rho.shape[:2]
    reg = []
    for t in range(shape_t):
        for _rho in torch.chunk(rho[t], n_chunks, dim=0):
            reg.append(grid_sample(
                pred[t].expand(_rho.shape[0], -1, -1, -1, -1),
                _rho.view(-1, 1, 1, 1, 3),
                align_corners=False, padding_mode='border'
            ))
    reg = torch.cat(reg, dim=0).view((shape_t, shape_n))
    return k * torch.mean(reg, dim=-1)



from logging import Logger
from pathlib import Path
from typing import List

from ..data import AnnotationTable, WorldlineTable
from ._utilities import default_args


@default_args("0")
def jump_to_frame(
    dataset: Path,
    annotations: AnnotationTable,
    worldlines: WorldlineTable,
    window_state,
    arg: str,
    logger: Logger
) -> List[dict]:
    """
    Jump to the specified frame (starting with 0).

    Args: frame_number

        frame_number: [default 0]
    """

    arg_list = arg.replace(" ", "").split(",")
    t = int(arg_list[0])

    return [
        {
            "type": "annotation_window/set_t_idx",
            "payload": t
        },
    ]



from logging import Logger
from pathlib import Path
from typing import List

from ..data import AnnotationTable, WorldlineTable, Annotation

def insert_annotation(
    dataset: Path,
    annotations: AnnotationTable,
    worldlines: WorldlineTable,
    window_state,
    arg: str,
    logger: Logger
) -> List[dict]:
    """
    Insert an annotation at the current coordinates using the current track.
    """

    logger.info("arg: {}".format(arg))

    A = Annotation(
        t_idx=window_state["t_idx"],
        x=window_state["x"],
        y=window_state["y"],
        z=window_state["z"],
        worldline_id=window_state["selected_worldline"] or 0,
        parent_id=0,
        provenance=window_state["active_provenance"]
    )
    a = annotations.insert(A)

    return [{
            "type": "annotations/get_annotations"
        },
        {
            "type": "annotation_window/set_selected_annotation_local",
            "payload": int(a.id)
        }]



"""
ZephOD network uses all functions listed below as input channels to combine/select from.
This file may be edited by a user to fit their particular use case,
or to add new State-of-the-Art segmentation/detection algorithms,
but each edit will require retraining of the weights.
Each function needs to take only the volumetric data slice (Z, Y, X) as the input,
and return the processed volume with the same shape as the output.
Each function will be iterated over all data channels.
"""

import cv2
import numpy as np
from skimage import restoration

from ..utils.utils import gaussian


def rl_deconvolution(img, sigma=5, gamma=1.0, lam=12, n_iter=10):
    processed = np.zeros_like(img)
    kernel = gaussian(
        np.array([1, sigma, sigma]),
        shape=None, dtype=np.float32, norm="max"
    )
    for z in range(img.shape[0]):
        gray = (img[z].astype('float64') / (np.max(img[z]) + 1E-8)) ** (1 / gamma)
        image_noisy = gray + (np.random.poisson(lam=lam, size=gray.shape) - 10) / 255
        processed[z] = restoration.richardson_lucy(image_noisy, kernel[1], iterations=n_iter)

    return processed * 255


def apply_255_lut(img: np.ndarray, lo=0, hi=255) -> np.ndarray:
    if lo == 0:
        lo = max(np.amin(img), 0.1 * np.amax(img))
    if hi == 255:
        hi = max(np.amax(img), 1.0)

    newtype = img.dtype

    y_float = (img - lo) / (hi - lo)
    y_clipped = np.clip(y_float, 0, 255)

    if np.issubdtype(newtype, np.integer):
        maxval = np.iinfo(newtype).max
    else:
        maxval = 255.0

    return (maxval * y_clipped).astype(newtype)


# Img must be of shape - (Z,Y,X) (channels not included)
def threshold(img: np.ndarray, lo=50, hi=200) -> np.ndarray:
    processed = np.zeros_like(img)
    for z in range(img.shape[0]):
        ret, thresh = cv2.threshold(img[z], lo, hi, cv2.THRESH_BINARY)
        kernel = np.ones((1, 2), np.uint8)
        processed[z] = cv2.erode(thresh, kernel, iterations=1)
    return processed


# Img must be of shape - (Z,Y,X) (channels not included)
def sharpen(img: np.ndarray,
            sharp=np.array([[0, -3, 0], [-3, 16, -3], [0, -3, 0]])
            ) -> np.ndarray:
    processed = np.zeros_like(img)
    for z in range(img.shape[0]):
        img_lut = apply_255_lut(img[z])
        processed[z] = cv2.filter2D(img_lut, -1, sharp)
    return processed



from .build_annotations import build_annotations
from .build_models import build_models
from .build_springs import build_springs
from .build_tree import build_tree
from .track_all import track_all
from .save_annotations import save_annotations
from .save_movie import save_movie



from ..utils.utils import *


def build_annotations(
    container,
    annotation,
    t_ref,
    wlid_ref,
    n_ref,):
    """Load and handle annotations from annotations.h5 file.

    Annotations are loaded and sorted according to user arguments, and used to
    populate an empty results array. This array will be filled during tracking.

    :param container: variable container; needs to contain: dataset,
    exclude_self, exclusive_prov, shape_t
    :param annotation: override annotations to use instead of loading from file
    :param t_ref: override frames to use as annotations
    :param wlid_ref: override worldline id's to analyze
    :param n_ref: override maximum number of keypoints to analyze
    :return: container (updated entries for: annot, shape_n, partial_annot,
    provenance, t_annot, worldline_id), results (pre-filled with loaded annotations)
    """

    # pull variables from container
    dataset = container.get('dataset')
    exclude_self = container.get('exclude_self')
    exclusive_prov = container.get('exclusive_prov')
    shape_t = container.get('shape_t')

    # checking annotated frames
    if annotation is None:
        annotation = get_annotation_df(dataset)
    t_annot = np.unique(annotation['t_idx']).astype(int)
    if t_ref is not None:
        if type(t_ref) is int:
            t_ref = [t_ref]
        t_annot = np.array([t for t in t_ref if t in t_annot]).astype(int)
    t_annot = np.sort(t_annot)

    worldline_id = None
    if wlid_ref is not None:
        if type(wlid_ref) is int:
            worldline_id = np.arange(wlid_ref)
        elif type(wlid_ref) is tuple and len(wlid_ref) == 2:
            worldline_id = np.arange(min(wlid_ref), max(wlid_ref))
        elif type(wlid_ref) is tuple or type(wlid_ref) is list:
            worldline_id = np.sort(np.array(wlid_ref))
        shape_n = len(worldline_id)
    elif n_ref is not None:
        shape_n = n_ref
        for t in t_annot:
            u, _, _ = get_annotation(annotation, t, exclusive_prov, exclude_self)
            if len(u) == n_ref:
                worldline_id = u
                print(f'Using frame #{t} as initial reference with specified {n_ref} annotations...')
                break
    else:
        nn_list = [len(get_annotation(annotation, t, exclusive_prov, exclude_self)[0]) for t in t_annot]
        shape_n, t_max = np.max(nn_list), t_annot[np.argmax(nn_list)]
        worldline_id, _, _ = get_annotation(annotation, t_max, exclusive_prov, exclude_self)
        print(f'Using frame #{t_max} as initial reference with {shape_n} annotations found...')

    if shape_n == 0 or shape_n is None or worldline_id is None:
        print('\n******* ERROR: annotations could not be loaded properly! '
              'Check parameters: t_ref, wlid_ref, n_ref.\n\n')

    annot = []
    partial_annot = {}
    results = np.zeros((shape_t, shape_n, 3))
    provenance = np.array([[b'ZEIR'] * shape_n] * shape_t)

    for t in t_annot:
        # loading and sorting annotation by worldline_id
        u, _annot, prov = get_annotation(annotation, t, exclusive_prov, exclude_self)

        # checking if worldlines are available in the annotation
        w_idx = np.array([np.where(worldline_id == w)[0][-1]
                          for w in u if w in worldline_id], dtype=int)
        u_idx = np.array([np.where(u == w)[0][-1]
                          for w in worldline_id if w in u], dtype=int)
        _annot = _annot[u_idx, ...]

        if _annot.shape[0] > shape_n or _annot.shape[0] == 0:
            t_annot = np.setdiff1d(t_annot, [t])
            continue
        elif _annot.shape[0] < shape_n:
            t_annot = np.setdiff1d(t_annot, [t])
            partial_annot[t] = (w_idx, _annot)
            results[t, w_idx] = _annot
            provenance[t, w_idx] = prov[u_idx]
            continue

        annot.append(_annot)
        results[t] = _annot
        provenance[t, w_idx] = prov[u_idx]

    print(f'\nAnnotations loaded for frames {list(t_annot)} '
          f'with shape: {np.array(annot).shape}')
    if len(partial_annot) > 0:
        print(f'*** Partial annotations found for {len(partial_annot)} frames: '
              f'{list(partial_annot.keys())}')

    # push variables to container
    container.update({
        'annot': annot,
        'shape_n': shape_n,
        'partial_annot': partial_annot,
        'provenance': provenance,
        't_annot': t_annot,
        'worldline_id': worldline_id,
    })

    # push results to checkpoint
    update_checkpoint(dataset, {
        'results': results,
    })

    return container, results



"""
ZephIR: multiple object tracking via image registration

See /docs/Guide-parameters.md for detailed explanations and tips for using optional arguments.

Usage:
    zephir -h | --help
    zephir -v | --version
    zephir --dataset=<dataset> [options]
    zephir --dataset=<dataset> [options]

Options:
    -h --help                           show this message and exit.
    -v --version                        show version information and exit.
    --dataset=<dataset>                 path to data directory to analyze.
    --load_checkpoint=<load_checkpoint>  resume from last checkpoint. [default: False]
    --load_args=<load_args>             load arguments from existing args.json file. [default: False]
    --allow_rotation=<allow_rotation>   enable rho parameter to rotate descriptors. [default: False]
    --channel=<channel>                 data channel to register.
    --clip_grad=<clip_grad>             maximum value for gradients; use -1 to uncap gradients. [default: 1.0]
    --cuda=<cuda>                       check if a CUDA-compatible GPU is available for  use. [default: True]
    --dimmer_ratio=<dimmer_ratio>       ratio to dim out non-foveated regions. [default: 0.1]
    --exclude_self=<excluse_self>       exclude annotations with provenance 'NEIR'. [default: True]
    --exclusive_prov=<exclusive_prov>   only load annotations with given provenance.
    --fovea_sigma=<fovea_sigma>         sigma for gaussian mask for foveated regions; use -1 to disable. [default: 2.5]
    --gamma=<gamma>                     gamma correction coefficient. [default: 2]
    --grid_shape=<grid_shape>           size of image descriptor in the xy-plane. [default: 25]
    --include_all=<include_all>         include all existing annotations in save file. [default: True]
    --lambda_d=<lambda_d>               coefficient for center detection loss. [default: -1.0]
    --lambda_n=<lambda_n>               coefficient for intra-keypoint spring constant. [default: 1.0]
    --lambda_n_mode=<lambda_n_mode>     method by which spring loss is calculated. [default: disp]
    --lambda_t=<lambda_t>               coefficient for temporal loss. [default: -1.0]
    --load_nn=<load_nn>                 load in manually defined spring connections if available. [default: True]
    --lr_ceiling=<lr_ceiling>           maximum value for initial learning rate. [default: 0.2]
    --lr_coef=<lr_coef>                 coefficient for initial learning rate. [default: 2.0]
    --lr_floor=<lr_floor>               minimum value for initial learning rate. [default: 0.02]
    --motion_predict=<motion_predict>   enable flow field motion prediction with partial annotations. [default: False]
    --n_chunks=<n_chunks>               number of steps to divide the forward pass into. [default: 10]
    --n_epoch=<n_epoch>                 number of iterations. [default: 40]
    --n_epoch_d=<n_epoch_d>             number of iterations for center detection. [default: 10]
    --n_frame=<n_frame>                 number of frames to analyze for temporal loss. [default: 1]
    --n_ref=<n_ref>                     override for shape_n; requires an annotated frames with exactly n_ref keypoints.
    --nn_max=<nn_max>                   maximum number of neighbors connected for spring loss. [default: 5]
    --save_mode=<save_mode>             mode for saving results. [default: o]
    --sort_mode=<sort_mode>             method for sorting frame order and parent-child tree. [default: similarity]
    --t_ignore=<t_ignore>               frames to ignore for analysis.
    --t_ref=<t_ref>                     override for reference frames to search for annotations.
    --t_track=<t_track>                 frames to include for analysis; supercedes t_ignore.
    --wlid_ref=<wlid_ref>               subset of keypoints to track by worldline_id; supercedes n_ref.
    --z_compensator=<z_compensator>     number of additional gradient descent steps for z-axis. [default: -1.0]
"""

import shutil
from docopt import docopt

from .__version__ import __version__
from .methods import *
from .models.container import Container
from .utils.io import *


def run_zephir(dataset: Path, args: dict):

    if not (dataset / 'backup').is_dir():
        Path.mkdir(dataset / 'backup')

    if args['--load_checkpoint'] in ['True', 'Y', 'y']:
        args = get_checkpoint(dataset, 'args', verbose=True)
        state = get_checkpoint(dataset, 'state')
        if args is None or state is None:
            print('*** CHECKPOINT EMPTY! Exiting...')
            exit()
    else:
        if (
            args['--load_args'] in ['True', 'Y', 'y'] and
            (dataset / 'args.json').is_file()
        ):
            with open(str(dataset / 'args.json')) as json_file:
                args = json.load(json_file)

        update_checkpoint(
            dataset,
            {'state': 'init',
             '__version__': __version__,
             'args': args}
        )
        state = 'init'

    # checking for available CUDA GPU
    if args['--cuda'] in ['True', 'Y', 'y'] and torch.cuda.is_available():
        print('\n*** GPU available!')
        dev = 'cuda'
    # elif torch.backends.mps.is_available():
    #     dev = 'mps'
    else:
        dev = 'cpu'
    print(f'\nUsing device: {dev}\n\n')

    # building/loading variable container
    if state == 'init':

        container = Container(
            dataset=dataset,
            allow_rotation=args['--allow_rotation'] in ['True', 'Y', 'y'],
            channel=int(args['--channel']) if args['--channel'] else None,
            dev=dev,
            exclude_self=args['--exclude_self'] in ['True', 'Y', 'y'],
            exclusive_prov=(bytes(args['--exclusive_prov'], 'utf-8')
                            if args['--exclusive_prov'] else None),
            gamma=float(args['--gamma']),
            include_all=args['--include_all'] in ['True', 'Y', 'y'],
            lr_coef=float(args['--lr_coef']),
            n_frame=int(args['--n_frame']),
            z_compensator=float(args['--z_compensator']),
        )

        update_checkpoint(dataset, {'state': 'load'})
        state = 'load'

    else:
        container = get_checkpoint(dataset, 'container')

    # building annotations table and tracking models
    if state == 'load':

        container, results = build_annotations(
            container=container,
            annotation=None,
            t_ref=eval(args['--t_ref']) if args['--t_ref'] else None,
            wlid_ref=eval(args['--wlid_ref']) if args['--wlid_ref'] else None,
            n_ref=int(args['--n_ref']) if args['--n_ref'] else None,
        )

        update_checkpoint(dataset, {'state': 'build'})
        state = 'build'

    else:
        results = get_checkpoint(dataset, 'results')

    # compiling spring network and frame tree
    if state == 'build':

        container, zephir, zephod = build_models(
            container=container,
            dimmer_ratio=float(args['--dimmer_ratio']),
            grid_shape=(5, 2 * (int(args['--grid_shape']) // 2) + 1,
                        2 * (int(args['--grid_shape']) // 2) + 1),
            fovea_sigma=(1, float(args['--fovea_sigma']),
                         float(args['--fovea_sigma'])),
            n_chunks=int(args['--n_chunks']),
        )

        container = build_springs(
            container=container,
            load_nn=args['--load_nn'] in ['True', 'Y', 'y'],
            nn_max=int(args['--nn_max']),
        )

        container = build_tree(
            container=container,
            sort_mode=str(args['--sort_mode']),
            t_ignore=eval(args['--t_ignore']) if args['--t_ignore'] else None,
            t_track=eval(args['--t_track']) if args['--t_track'] else None,
        )

        update_checkpoint(dataset, {'state': 'track', '_t_list': None})
        state = 'track'

    else:
        zephir = get_checkpoint(dataset, 'zephir')
        zephod = get_checkpoint(dataset, 'zephod')

    # tracking all frames in _t_list
    if state == 'track':

        container, results = track_all(
            container=container,
            results=results,
            zephir=zephir,
            zephod=zephod,
            clip_grad=float(args['--clip_grad']),
            lambda_t=float(args['--lambda_t']),
            lambda_d=float(args['--lambda_d']),
            lambda_n=float(args['--lambda_n']),
            lambda_n_mode=args['--lambda_n_mode'],
            lr_ceiling=float(args['--lr_ceiling']),
            lr_floor=float(args['--lr_floor']),
            motion_predict=args['--motion_predict'] in ['True', 'Y', 'y'],
            n_epoch=int(args['--n_epoch']),
            n_epoch_d=(int(args['--n_epoch_d'])
                       if float(args['--lambda_d']) > 0 else 0),
            _t_list=get_checkpoint(dataset, '_t_list'),
        )

    else:
        results = get_checkpoint(dataset, 'results', verbose=True)

    if np.any(np.isnan(results)):
        print(f'*** WARNING: NaN found in: '
              f'{list(np.unique(np.where(np.isnan(results))[0]))}')
        results = np.where(np.isfinite(results), results, 0)
        update_checkpoint(dataset, {'results': results})

    save_annotations(
        container=container,
        results=results,
        save_mode=str(args['--save_mode']),
    )

    save_movie(
        container=container,
        results=results,
    )

    now = datetime.datetime.now()
    now_ = now.strftime("%m_%d_%Y_%H_%M_%S")
    shutil.copy(dataset / 'checkpoint.pt',
                dataset / 'backup' / f'checkpoint_{now_}.pt')

    print('\n\n*** DONE!')
    return


def main():
    args = docopt(__doc__, version=f'ZephIR {__version__}')
    # print(args, '\n')

    dataset = Path(args['--dataset'])

    run_zephir(
        dataset=dataset,
        args=args
    )


if __name__ == '__main__':
    main()



# Flask server to provide read access to image volumes and read/write access
# to annotations stored in a provided path.
#
# This file can be run in two ways:
#  1. For development, use flask run annotator.py
#  2. For production, use annotator [OPTIONS]
"""
This provides a webpage to label key-points in a given dataset.

Usage:
    annotator [options]

Options:
    -h --help                   Show this help.
    -v --version                Show version information.
    --dataset=PATH              Location of data for labeling.
    --port=PORT                 Port to serve on. [default: 5000]
"""

from functools import lru_cache
import io
import os
from pathlib import Path
import subprocess
import traceback
from typing import Tuple

from docopt import docopt
from flask import (Flask, jsonify, send_file, request, send_from_directory)
from gevent.pywsgi import WSGIServer
import numpy as np
from skimage.measure import block_reduce

from .data import (get_slice, get_metadata, Annotation,
                   Worldline, load_annotations, save_annotations,
                   stash_annotations, mip_x, mip_y, mip_z)

from . import rpc as rpc
from ..__version__ import __version__


app = Flask(__name__)
app.logger.setLevel("INFO")

app.dataset = Path(r"D:\data\201124")

(app.annotations, app.worldlines) = load_annotations(app.dataset)


def get_volume(t: int) -> np.ndarray:
    return get_slice(app.dataset, t)


@lru_cache(maxsize=100)
def get_rgba_volume(t: int) -> np.ndarray:
    """Add blank blue and alpha channels to an array with max-stride color
    dimension, permute the color dimension to the min-stride position, and
    rearrange memory to make the color dimension min-stride (last)."""

    raw_vol = get_volume(t)

    if raw_vol.ndim == 4:

        shape_c = raw_vol.shape[0]
        shape_zyx = raw_vol.shape[1:]

        if shape_c == 1:
            g = np.zeros((1, *shape_zyx), dtype=np.uint8)
            b = np.zeros((1, *shape_zyx), dtype=np.uint8)
            rgb = np.concatenate((raw_vol, g, b))

        elif shape_c == 2:
            b = np.zeros((1, *shape_zyx), dtype=np.uint8)
            rgb = np.concatenate((raw_vol, b))

        else:
            rgb = raw_vol[:3, ...]

    else:
        r = np.expand_dims(raw_vol, 0)
        shape_zyx = r.shape[1:]
        g = np.zeros((1, *shape_zyx), dtype=np.uint8)
        b = np.zeros((1, *shape_zyx), dtype=np.uint8)
        rgb = np.concatenate((r, g, b))

    alpha = 255 * np.ones((1, *shape_zyx), dtype=np.uint8)

    rgba_czyx = np.concatenate((rgb, alpha))

    rgba_zyxc = rgba_czyx.transpose((1, 2, 3, 0)).copy()

    return rgba_zyxc


# Define the 2D views of our volume to paint on a screen. HTML canvas expects
# a 3D array (yxc), where the "c" is 4 bytes (rgb-alpha) with the shortest
# stride. These should just return the bytes to draw on a canvas.


def get_z(t: int, idx: int) -> bytes:
    zyxc = get_rgba_volume(t)
    yxc = zyxc[idx]
    return yxc.tobytes()


def get_y(t: int, idx: int) -> bytes:
    zyxc = get_rgba_volume(t)
    zxc = zyxc[:, idx, :, :]
    return zxc.tobytes()


def get_x(t: int, idx: int) -> bytes:
    zyxc = get_rgba_volume(t)
    zyc = zyxc[:, :, idx, :]
    yzc = zyc.transpose((1, 0, 2))
    return yzc.tobytes()


def get_mip_z(t: int, idx: int) -> bytes:
    return np.ascontiguousarray(mip_z(get_rgba_volume(t)))


def get_mip_y(t: int, idx: int) -> bytes:
    return np.ascontiguousarray(mip_y(get_rgba_volume(t)))


def get_mip_x(t: int, idx: int) -> bytes:
    return np.ascontiguousarray(mip_x(get_rgba_volume(t)))


def downsample(v: np.ndarray, scale: Tuple[int, ...]) -> np.ndarray:
    """Downsample a volume by a factor that divides the given dimension."""
    #mean8 = lambda x, **kwargs: np.mean(x, dtype=v.dtype, **kwargs)
    return block_reduce(v, scale, np.max)


def get_vol(t: int, idx: int) -> bytes:
    zyxc = get_rgba_volume(t)
    return downsample(zyxc, (1, idx, idx, 1)).tobytes()


def get_gvol(t: int, idx: int) -> bytes:
    """Just the green channel as an 8-bit array."""
    zyxc = get_rgba_volume(t)
    zyxG = zyxc[:, :, :, 1]
    zyxG_scaled = downsample(zyxG, (1, idx, idx))
    #zyxG_scaled = resize(zyxG_scaled, (1, idx, idx))
    return zyxG_scaled.tobytes()


# Serve the static client files


@app.route('/', methods=['GET'])
def serve_dir_directory_index():
    return send_from_directory("client_build", "index.html")


@app.route('/favicon.ico', methods=['GET'])
def serve_favicon():
    return send_from_directory("client_build", "favicon.ico")


@app.route('/manifest.json', methods=['GET'])
def serve_manifest():
    return send_from_directory("client_build", "manifest.json")


@app.route('/static/js/<path:path>', methods=['GET'])
def serve_static_js(path):
    return send_from_directory(os.path.join("client_build", "static", "js"),
                               path)


@app.route('/static/css/<path:path>', methods=['GET'])
def serve_static_css(path):
    return send_from_directory(os.path.join("client_build", "static", "css"),
                               path)


@app.route('/webgl_colormaps/<path:path>', methods=['GET'])
def serve_colormaps(path):
    return send_from_directory(os.path.join("client_build", "webgl_colormaps"),
                               path)


# End static files.


@app.route("/metadata")
def get_app_metadata():
    metadata = get_metadata(app.dataset)
    return jsonify(metadata)


@app.route("/t/<t>/data/<view>/<view_idx>")
def get_binary_image(t, view, view_idx):

    t = int(t)
    view_fn = "get_{}".format(view)
    view_idx = int(view_idx)

    bytes_to_send = globals()[view_fn](t, view_idx)
    buffer = io.BytesIO(bytes_to_send)

    return send_file(buffer,
                     as_attachment=True,
                     attachment_filename="data.bin",
                     mimetype="application/octet-stream")


@app.route("/t/<t>/annotations", methods=["GET"])
def get_annotations(t):

    t = int(t)
    annotations_for_t = app.annotations.get_t(t).to_jsonable_dict()

    return jsonify(annotations_for_t)


@app.route("/annotations", methods=["POST"])
def handle_annotations():

    app.logger.info("Inserting annotation {}".format(request.json))
    a = Annotation(**request.json)
    inserted_a = app.annotations.insert(a)

    return jsonify(inserted_a.to_jsonable_dict())


@app.route("/annotations/<id>", methods=["PUT", "DELETE"])
def handle_annotation(id):

    id = int(id)

    if request.method == "PUT":

        app.logger.info("Updating annotation {} to {}".format(
            id, request.json))

        try:

            updated_a = app.annotations.update(id, request.json)

        except Exception as e:
            app.logger.warn(
                "Update failed. Returning original. Error: {}".format(e))
            updated_a = app.annotations.get(id)

        return jsonify(updated_a.to_jsonable_dict())

    elif request.method == "DELETE":

        try:
            app.logger.info("Deleting annotation {}".format(id))
            deleted_id = app.annotations.delete(id)
            if deleted_id == id:
                return "ok"
            else:
                return "error"

        except:
            return "error"

    else:

        app.logger.info(f"Received invalid request: {request.method}")
        return "error"


@app.route("/worldlines", methods=["GET", "POST"])
def handle_worldlines():

    if request.method == "GET":

        app.logger.info("Fetching worldlines.")
        return jsonify(app.worldlines.to_jsonable_dict())

    elif request.method == "POST":

        app.logger.info("Creating a new worldline.")
        w = app.worldlines.insert(Worldline())
        return jsonify(w.to_jsonable_dict())

    else:

        app.logger.info(f"Received invalid request: {request.method}")
        return "error"


@app.route("/worldlines/<id>", methods=["PUT", "DELETE"])
def handle_worldline(id):

    id = int(id)

    if request.method == "PUT":

        app.logger.info("Updating worldline {}: {}".format(id, request.json))
        w = app.worldlines.update(id, request.json)
        app.logger.info("Updated worldline {}: {}".format(
            id, w.to_jsonable_dict()))
        return jsonify(w.to_jsonable_dict())

    elif request.method == "DELETE":

        app.logger.info("Deleting worldline {}".format(id))
        deleted_id = app.worldlines.delete(id)
        if deleted_id == id:
            return "ok"
        else:
            return "error"

    else:

        app.logger.info(f"Received invalid request: {request.method}")
        return "error"


@app.route("/load", methods=["POST"])
def handle_load():

    app.logger.info("Loading all annotations and worldlines from disk.")
    stash_annotations(app.annotations, app.worldlines, app.dataset)
    (app.annotations, app.worldlines) = load_annotations(app.dataset)

    return jsonify({"path": str(app.dataset)})


@app.route("/save", methods=["POST"])
def handle_save():

    app.logger.info("Saving all annotations and worldlines to disk.")
    save_annotations(app.annotations, app.worldlines, app.dataset)
    return jsonify({"path": str(app.dataset)})


# RPCs can alter the annotations and worldlines states. They are provided with
# the state of the annotation window, the AnnotationTable, and the
# WorldlinesTable. They can return a list of actions to dispatch, which should
# allow the application to get all the updates that were made.


@app.route("/rpc", methods=["GET", "POST"])
def handle_rpc():

    if request.method == "GET":

        listing = rpc.__all__
        app.logger.info("rpc list: {}".format(listing))

        rpc_list = []
        for proc in listing:
            rpc_list.append({
                "name":
                proc,
                "default_args":
                str(rpc.__dict__.get(proc).__dict__.get("default_args", "")),
                "docstring":
                rpc.__dict__.get(proc).__doc__ or ""
            })

        payload = {"status": "OK", "payload": rpc_list}
        return jsonify(payload)

    elif request.method == "POST":

        data = request.json
        app.logger.info("Received rpc request: {}".format(data["method"]))

        proc = getattr(rpc, data["method"], None)

        try:
            actions = proc(app.dataset, app.annotations, app.worldlines,
                           data["state"], data["arg"], app.logger)

            return jsonify({"status": "ok", "callbacks": actions})

        except Exception as e:
            return jsonify({
                "status": "error",
                "exception": str(e),
                "traceback": traceback.format_exc()
            })

    else:

        app.logger.info(f"Received invalid request: {request.method}")
        return "error"


def main():
    """CLI entry point."""

    args = docopt(__doc__, version=f'ZephIR annotator {__version__}')

    if args["--dataset"]:
        app.dataset = Path(args["--dataset"])
        (app.annotations, app.worldlines) = load_annotations(app.dataset)

    port = int(args["--port"])

    print("Starting a server on port {}".format(port))

    http_server = WSGIServer(("0.0.0.0", port), app)
    http_server.serve_forever()


if __name__ == "__main__":
    main()



"""
This file collects key functions for preprocessing data and annotations.
"""

import datetime
from functools import lru_cache, partial
import h5py
import importlib.util
import json
import numpy as np
import pathlib
from pathlib import Path
import platform
import torch
from tqdm import tqdm
from typing import Optional, Callable
from types import ModuleType


# from .getters import *
from . import getters as default_getters
from ..__version__ import __version__


def _module_name_from_path(p: Path) -> str:
    """Return a globally unique module name associated with a given
    directory."""
    return str(p / "getters")


def _getter_module_from_path(dataset: Path) -> Optional[ModuleType]:
    module_name = _module_name_from_path(dataset)
    getter_file = dataset / "getters.py"

    if not getter_file.is_file():
        return None

    spec = importlib.util.spec_from_file_location(module_name, getter_file)

    getters = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(getters)  # type: ignore

    return getters


def _make_getter_with_default(name: str, default: Callable) -> Callable:
    """This hideous function creates a getter with the given name. If that
    function doesn't exist in a "getter.py" file within the dataset, then
    a specified default function is called. The default is bound with
    the path."""

    def _getter(p: Path, *getter_args):
        getters = _getter_module_from_path(p)

        if getters is None or not hasattr(getters, name):
            _fn = partial(default, p)

        else:
            _fn = partial(getters.__getattribute__(name), p)

        return _fn(*getter_args)

    _getter.__name__ = name

    return _getter


get_slice = _make_getter_with_default('get_slice', default_getters.get_slice)
get_metadata = _make_getter_with_default('get_metadata', default_getters.get_metadata)
get_annotation_df = _make_getter_with_default('get_annotation_df', default_getters.get_annotation_df)


@lru_cache()
def get_data(dataset, t, g=1, c=None):
    data = get_slice(dataset, t).astype(float)
    data = np.power(data / np.max(data), 1/g)
    if len(data.shape) == 3:
        return data[np.newaxis, ...]
    elif c is None:
        return data
    elif c >= 0:
        return data[c, np.newaxis, ...]
    elif c == -1:
        return np.max(data, axis=0)[np.newaxis, ...]
    return data


def get_annotation(annotation, t, exclusive_prov=None, exclude_self=True):
    annot = np.stack(
        [(annotation['x'][annotation['t_idx'] == t]).astype(float) * 2 - 1,
         (annotation['y'][annotation['t_idx'] == t]).astype(float) * 2 - 1,
         (annotation['z'][annotation['t_idx'] == t]).astype(float) * 2 - 1
         ], axis=-1
    )
    worldline_id = annotation['worldline_id'][annotation['t_idx'] == t]
    provenance = np.array(annotation['provenance'][annotation['t_idx'] == t])
    if exclusive_prov is not None:
        annot = annot[provenance == exclusive_prov, :]
        worldline_id = worldline_id[provenance == exclusive_prov]
        provenance = provenance[provenance == exclusive_prov]
    elif exclude_self:
        annot = annot[provenance != b'ZEIR', :]
        worldline_id = worldline_id[provenance != b'ZEIR']
        provenance = provenance[provenance != b'ZEIR']

    u, i, c = np.unique(worldline_id, return_index=True, return_counts=True)
    ovc_idx = np.where(c > 1)[0]
    for j in ovc_idx:
        i[j] = np.where(worldline_id == u[j])[0][-1]
    return u, annot[i, ...], provenance[i]


def get_checkpoint(path, key: str, verbose=False):
    """
    Retrieves given key and value from an existing checkpoint.
    """

    checkpoint = load_checkpoint(path, verbose=verbose)

    if key in checkpoint.keys():
        value = checkpoint[key]
        del checkpoint
        return value
    elif verbose:
        print(f'*** ERROR: {key} not found in checkpoint!')
    del checkpoint
    return None


def update_checkpoint(path, dictionary: dict, verbose=True):
    """
    Updates an existing checkpoint dictionary with given key and value and saves to file.
    """

    checkpoint = load_checkpoint(path, verbose=False)

    for key in dictionary.keys():
        checkpoint[key] = dictionary[key]

        if key == 'args':
            with open(Path(path) / 'args.json', 'w') as outfile:
                json.dump(dictionary[key], outfile, indent=4)

    checkpoint['last_update'] = str(datetime.datetime.now())

    with open(Path(path) / 'checkpoint.pt', 'wb') as f:
        torch.save(checkpoint, f)

    if verbose:
        print(f'Checkpoint updated for {list(dictionary.keys())} '
              f'@ [{checkpoint["last_update"]}]')

    del checkpoint
    return


def load_checkpoint(path, fallback=True, verbose=True):
    if verbose:
        print('Loading checkpoint...')

    file_name = Path(path) / 'checkpoint.pt'
    checkpoint, map_loc = None, None
    if file_name.is_file():

        while checkpoint is None:

            try:
                checkpoint = torch.load(str(file_name), map_location=map_loc)

            except ModuleNotFoundError:
                print('*** ERROR: checkpoint not compatible with current ZephIR version!')
                if fallback:
                    print('Creating empty checkpoint...')
                    checkpoint = {'__version__': __version__}
                else:
                    exit()

            except (NotImplementedError, RuntimeError) as e:
                if platform.system() == 'Darwin' or platform.system() == 'Linux':
                    pathlib.WindowsPath = pathlib.PosixPath
                elif platform.system() == 'Windows':
                    pathlib.PosixPath = pathlib.WindowsPath

            except RuntimeError:
                print('*** CUDA NOT AVAILABLE! Mapping CUDA tensors to CPU...')
                map_loc = torch.device('cpu')

    elif fallback:
        print('*** CHECKPOINT NOT FOUND! Creating empty checkpoint...')
        return {'__version__': __version__}

    else:
        print('*** CHECKPOINT NOT FOUND!')
        exit()

    if verbose and '__version__' not in checkpoint.keys():
        print('*** WARNING: __version__ not found in checkpoint! '
              'Setting to current ZephIR version, but may not be accurate. '
              'Consider resetting checkpoint.\n')
        checkpoint['__version__'] = __version__
    elif verbose and __version__ != checkpoint['__version__']:
        print('*** WARNING: checkpoint does not match current ZephIR version! '
              'Consider resetting checkpoint.\n')
    elif verbose:
        print('Successfully loaded checkpoint!\n')

    return checkpoint


def update_metadata(path, dictionary: dict):
    """Update an existing metadata file with given key and value."""
    file_name = Path(path) / 'metadata.json'

    try:
        metadata = get_metadata(path)
    except (OSError, IOError) as e:
        print('*** WARNING: Error loading metadata!')
        if not file_name.is_file():
            print(f'Creating empty metadata\t'
                  f'-- properly populate this file with the right metadata!\n'
                  f'Saving to {file_name}...\n')
            metadata = {}
        else:
            exit()

    for key in dictionary.keys():
        metadata[key] = dictionary[key]

    with open(file_name, 'w') as outfile:
        json.dump(metadata, outfile, indent=4)



import cv2
import random

from inspect import getmembers, isfunction
from scipy.ndimage import rotate
from skimage.transform import resize
# from stardist.models import StarDist2D

from . import channels
from ..utils import utils
from ..utils.utils import *


def identify(vol, annotations, r_crop):
    if np.max(vol) <= 1.0:
        vol = (vol * 255).astype(np.uint8)

    isolates = []

    for annot in annotations:
        x, y, z = get_pixel(annot, vol.shape[1:])
        slice_yx = np.max(vol[0, max(0, z-1):z+2, y-r_crop:y+r_crop, x-r_crop:x+r_crop], axis=0)
        slice_zy = np.max(vol[0, max(0, z-2):z+3, y-r_crop:y+r_crop, x-1:x+2], axis=-1)
        # contours_yx = measure.find_contours(slice_yx, 0.8)
        # contours_zy = measure.find_contours(slice_zy, 0.8)
        contours_yx = cv2.findContours(slice_yx, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)[0]
        contours_zy = cv2.findContours(slice_zy, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)[0]
        if len(contours_yx) != 1 or len(contours_zy) != 1:
            continue

        crop = vol[:, max(0, z-2):z+3, y-r_crop:y+r_crop, x-r_crop:x+r_crop].copy()
        for channel in range(crop.shape[0]):
            for zslice in range(crop.shape[1]):
                if channel == 0:
                    _, thresh = cv2.threshold(crop[channel, zslice], 30, 255, cv2.THRESH_TOZERO)
                else:
                    _, thresh = cv2.threshold(crop[channel, zslice], 30, 255, cv2.THRESH_TOZERO)
                cntr = cv2.findContours(thresh, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)[0]
                mask = np.zeros_like(crop[channel, zslice])
                for c in cntr:
                    if cv2.pointPolygonTest(c, (mask.shape[0]//2, mask.shape[1]//2), False) > 0:
                        mask = cv2.drawContours(mask, [c], -1, 1, -1)
                        mask = cv2.drawContours(mask, [c], -1, 1, 3)
                crop[channel, zslice] = crop[channel, zslice] * mask

        if np.max(crop) > 0:
            isolates.append(crop)
            # plt.imshow(np.max(crop[0], axis=0))
            # plt.show()
    return isolates


def generate_synthetic_data(vol, annotations, isolates, preprocess=False, use_original=False):
    synthetic = np.zeros_like(vol)
    if use_original:
        synthetic = vol.copy()

    labels = np.zeros((1, *vol.shape[1:]))
    for n, annot in enumerate(annotations):
        x, y, z = get_pixel(annot, vol.shape[1:])

        if not use_original:
            neuron = isolates[random.randint(0, len(isolates)-1)]

            rotate_idx = random.randint(0, 1)
            if rotate_idx == 1:
                neuron = rotate_neuron(neuron)

            resize_idx = random.randint(0, 1)
            if resize_idx == 1:
                neuron = resize_neuron(neuron)

            reposition_idx = random.randint(0, 1)
            if reposition_idx == 1:
                x += random.randrange(-5, 5)
                y += random.randrange(-5, 5)
            #     z += random.randrange(-1, 1)

            while z-neuron.shape[1]//2 < 0:
                z += 1
            while z-neuron.shape[1]//2+neuron.shape[1] > synthetic.shape[1]:
                z += -1
            while y-neuron.shape[2]//2 < 0:
                y += 1
            while y-neuron.shape[2]//2+neuron.shape[2] > synthetic.shape[2]:
                y += -1
            while x-neuron.shape[3]//2 < 0:
                x += 1
            while x-neuron.shape[3]//2+neuron.shape[3] > synthetic.shape[3]:
                x += -1

            synthetic[
                :, z-neuron.shape[1]//2:z-neuron.shape[1]//2+neuron.shape[1],
                y-neuron.shape[2]//2:y-neuron.shape[2]//2+neuron.shape[2],
                x-neuron.shape[3]//2:x-neuron.shape[3]//2+neuron.shape[3]
            ] = np.max(
                np.append(
                    synthetic[
                        np.newaxis, :,
                        z-neuron.shape[1]//2:z-neuron.shape[1]//2+neuron.shape[1],
                        y-neuron.shape[2]//2:y-neuron.shape[2]//2+neuron.shape[2],
                        x-neuron.shape[3]//2:x-neuron.shape[3]//2+neuron.shape[3]
                    ], neuron[np.newaxis, ...],
                    axis=0
                ),
                axis=0
            )

        labels[:, z, y, x] = 1

    flip_idx = random.randint(-1, 2)
    if flip_idx < 2:
        for z in range(synthetic.shape[1]):
            for c in range(synthetic.shape[0]):
                synthetic[c][z] = cv2.flip(synthetic[c][z], flip_idx)
            labels[0][z] = cv2.flip(labels[0][z], flip_idx)

    if preprocess:
        for c in range(vol.shape[0]):
            synthetic = np.append(
                synthetic, np.stack(
                    [f[1](synthetic[c]) for f in getmembers(channels, isfunction)
                     if f not in getmembers(utils, isfunction)],
                    axis=0
                ), axis=0
            )
        for c in range(synthetic.shape[0]):
            synthetic[c] = synthetic[c] / (np.max(synthetic[c]) + 1E-8)

    return synthetic, labels


def rotate_neuron(neuron):
    angle = random.randrange(-40, 40)
    neuron = rotate(
        neuron, angle,
        axes=(3, 2), reshape=True, output=None,
        order=3, mode='constant', cval=0.0, prefilter=True
    )
    return neuron


def resize_neuron(neuron, shape=None):
    if shape is None:
        size = random.randrange(-3, 5)
        shape = (max(3, neuron.shape[2]+size), max(3, neuron.shape[3]+size))
    resized = np.zeros((*neuron.shape[:2], *shape))
    for channel in range(neuron.shape[0]):
        for zslice in range(neuron.shape[1]):
            resized[channel, zslice] = resize(
                neuron[channel, zslice],
                shape
            )
    return resized.astype(neuron.dtype)



# An annotation is a coordinate in a 4D array. Worldlines are lists of
# annotations.
#
# author: vivekv2@gmail.com

import base64
import colorsys
from dataclasses import dataclass, asdict, field
from functools import lru_cache, reduce
import json
from pathlib import Path
import uuid
from typing import Tuple, Optional

import h5py
import numpy as np
import pandas as pd

from .dataclass_helpers import DataclassTableBase
from ...utils.getters import get_annotation_df


def base64str_from_uint32(x: np.uint32):
    return base64.standard_b64encode(x.tobytes())[:6]


def get_random_base64_id():
    idu32 = np.uint32(uuid.uuid1().time_low)
    return base64str_from_uint32(idu32)


_S4 = np.dtype("S4")
_S7 = np.dtype("S7")


@dataclass
class Annotation:
    id: np.uint32 = 0  # 0 is unassigned. Get an id after inserting into table.
    t_idx: np.uint32 = 0  # in [0, shape_t)
    x: np.float32 = 0.5  # in (0, 1)
    y: np.float32 = 0.5  # in (0, 1)
    z: np.float32 = 0.5  # in (0, 1)
    worldline_id: np.uint32 = 0
    parent_id: np.uint32 = 0  # =0 is null/none
    provenance: _S4 = b"NULL"  # unknown

    def __post_init__(self):

        self.id = np.uint32(self.id)
        self.t_idx = np.uint32(self.t_idx)
        self.x = np.float32(self.x)
        self.y = np.float32(self.y)
        self.z = np.float32(self.z)
        self.worldline_id = np.uint32(self.worldline_id)
        self.parent_id = np.uint32(self.parent_id)
        self.provenance = np.string_(self.provenance)

        # if self.id == b"":
        #     self.id = get_random_base64_id()

    def to_tuple(self) -> tuple:
        return (self.id, self.t_idx, self.x, self.y, self.z, self.worldline_id,
                self.parent_id, self.provenance)

    def to_dict(self) -> dict:
        return asdict(self)

    def to_jsonable_dict(self) -> dict:

        jsonable_dict = {}
        for k, v in self.to_dict().items():
            if type(v) is bytes or type(v) is np.string_:
                v = v.decode()
            if type(v) is np.uint32:
                v = int(v)
            if type(v) is np.float32:
                v = str(v)
            jsonable_dict[k] = v

        return jsonable_dict


class AnnotationTable(DataclassTableBase):
    row_class = Annotation

    def get_t(self, t_idx: int):
        return self.filter(lambda x: x["t_idx"] == t_idx)

    def to_jsonable_dict(self) -> dict:

        jsonable_dict = {}
        for annotation in self:
            jsonable_dict[str(annotation.id)] = annotation.to_jsonable_dict()

        return jsonable_dict

    @staticmethod
    def from_hdf(filename: Path):

        # filename = Path(filename)
        #
        # f = h5py.File(filename, "r")
        # data = pd.DataFrame()
        #
        # # Handle 'correctly' shaped data from Pythons H5 library.
        # if len(f["id"].shape) == 1:
        #
        #     for k in f:
        #         data[k] = f[k]
        #
        #     return AnnotationTable(data)
        #
        # # Matlab cannot make 1D arrays, and it cannot save char arrays.
        # # Handle both of these here
        # else:
        #     for k in f:
        #         vals = np.squeeze(f[k])
        #         if k == "provenance":
        #             vals = np.array(
        #                 list(map(lambda x: x.tobytes(), vals)))
        #         data[k] = vals
        #
        #     return AnnotationTable(data)

        return AnnotationTable(get_annotation_df(Path(filename)))


@dataclass
class Worldline:
    id: np.uint32 = 0
    name: np.string_ = b"null"
    color: _S7 = b"#ffffff"

    def __post_init__(self):

        self.name = np.string_(self.name)
        self.color = np.string_(self.color)

    def to_dict(self) -> dict:
        return asdict(self)

    def to_jsonable_dict(self):

        jsonable_dict = {}
        for k, v in self.to_dict().items():
            if type(v) is bytes or type(v) is np.string_:
                v = v.decode()
            if type(v) is np.uint32:
                v = int(v)
            if type(v) is np.float32:
                v = str(v)
            jsonable_dict[k] = v

        return jsonable_dict


class WorldlineTable(DataclassTableBase):
    row_class = Worldline

    def to_jsonable_dict(self) -> dict:

        jsonable_dict = {}
        for worldline in self:
            jsonable_dict[str(worldline.id)] = worldline.to_jsonable_dict()

        return jsonable_dict

    @staticmethod
    def from_annotations(annotations: AnnotationTable):

        worldlines = WorldlineTable()
        worldlines._insert_and_preserve_id(Worldline(id=0))

        worldline_ids = np.unique(annotations.df["worldline_id"])
        for id in worldline_ids:
            worldlines.insert(Worldline(name=str(id)))

        return worldlines


@lru_cache()
def get_all_neuron_data() -> dict:

    file_path = Path(__file__).parent.absolute()
    neuron_list_file = file_path / "./neurons_celegans.json"
    print(file_path)
    with open(neuron_list_file) as fp:
        data = json.load(fp)
    return data


def get_neuron_data(x: str) -> dict:
    return get_all_neuron_data()["neurons"][x]


def cleanup_worldlines(A: AnnotationTable, W: WorldlineTable
                       ) -> Tuple[AnnotationTable, WorldlineTable]:
    """Sequentially renumber all annotations and worldlines, and remake the
    worldline table."""

    new_ids = range(1, len(A) + 1)
    old_ids = A.df.id
    update_aid = {old_ids[i]: new_ids[i] for i in range(len(new_ids))}
    update_aid[0] = 0

    A.df.id = A.df.id.apply(lambda x: update_aid[x])
    A.df.parent_id = A.df.parent_id.apply(lambda x: update_aid[x])

    used_W = np.unique(A.df.worldline_id)
    N = len(used_W)
    new_ids = range(1, N + 1)
    update_wid = {used_W[i]: new_ids[i] for i in range(N)}
    update_wid[0] = 0

    A.df.worldline_id = A.df.worldline_id.apply(lambda x: update_wid)

    W_new = WorldlineTable()
    W_new._insert_and_preserve_id(Worldline())

    for i in range(N):

        if used_W[i] in W.df.id:
            wline = W.get(used_W[i])
            wline.id = new_ids[i]
        else:
            wline = Worldline(id=new_ids[i])

        W_new._insert_and_preserve_id(wline)

    return (A, W_new)


def color_worldlines(W: WorldlineTable) -> None:
    """Overwrite the color of all worldlines using evenly spaced hues and
    random lightness / saturation, both high."""
    def _get_N_colors(N):
        colors = []
        for i in range(N):
            h = ((i * 157) % 360) / 360.
            l = (60 + 10 * np.random.rand()) / 100.
            s = (90 + 10 * np.random.rand()) / 100.
            rgb = colorsys.hls_to_rgb(h, l, s)
            rgb_256 = map(lambda x: max(0, min(int(x * 255), 255)), rgb)
            rgb_code = "#{0:02x}{1:02x}{2:02x}".format(*rgb_256)
            colors.append(rgb_code)
        return colors

    colors = _get_N_colors(len(W))
    W.df.color = colors


def load_annotations(dataset: Optional[Path] = None
                     ) -> Tuple[AnnotationTable, WorldlineTable]:

    if dataset is None:
        dataset = Path(".")

    annotation_file = dataset / "annotations.h5"

    if annotation_file.exists():
        annotations = AnnotationTable.from_hdf(dataset)
    else:
        annotations = AnnotationTable()

    worldline_file = dataset / "worldlines.h5"
    if worldline_file.exists():
        worldlines = WorldlineTable.from_hdf(worldline_file)
    else:
        worldlines = WorldlineTable.from_annotations(annotations)

    return (annotations, worldlines)


def save_annotations(annotations: AnnotationTable,
                     worldlines: WorldlineTable,
                     dataset: Path = None) -> None:

    if dataset is None:
        dataset = Path(".")

    annotations.to_hdf(dataset / "annotations.h5")
    worldlines.to_hdf(dataset / "worldlines.h5")


def stash_annotations(annotations: AnnotationTable,
                      worldlines: WorldlineTable,
                      dataset: Path = None) -> None:

    if dataset is None:
        dataset = Path(".")

    annotations.to_hdf(dataset / "annotations_unsaved.h5")
    worldlines.to_hdf(dataset / "worldlines_unsaved.h5")



from logging import Logger
from pathlib import Path
from typing import List

from ..data import (get_channel_specific_slice_3D, Annotation, Worldline,
                    AnnotationTable, WorldlineTable, get_nearby_max)
from ._utilities import default_args


@default_args("0.01, 0.08, 0.005, 0.04, False, _ACTIVE, *")
def insert_local_max(
    dataset: Path,
    annotations: AnnotationTable,
    worldlines: WorldlineTable,
    window_state,
    arg: str,
    logger: Logger
) -> List[dict]:
    """
    Insert an annotation at the local max near the current point.

    Args: r_x, r_z, blur_x, blur_z, new_worldline, provenance, channel

        r_x: Search radius in x, normalized units (0, 1). The radius in y is
            set by the aspect ratio.
        r_z: Search radius in z, normalized units (0, 1). It should be larger
            than r_x because of the typically smaller dimension.
        blur_x: The sigma of a gaussian blur to apply prior to finding the
            local maximum. blur_y is calculated using the aspect ratio.
        blur_z: Same for z.
        new_worldline: Create a new worldline/track for this annotation.
        provenance: Anything is okay. Special values:
            _ACTIVE: Use the currently active provenance [default]
        channel: channel to use to calculate local max:
            *: All channles [default]
    """

    aspect_ratio = window_state["shape_x"]/window_state["shape_y"]

    arg_list = arg.replace(" ", "").split(",")
    r_x = float(arg_list[0])
    r_y = r_x * aspect_ratio
    r_z = float(arg_list[1])
    blur_x = float(arg_list[2])
    blur_y = blur_x * aspect_ratio
    blur_z = float(arg_list[3])
    new_worldline = arg_list[4] == "True"
    provenance = arg_list[5]
    channel = arg_list[6]

    if provenance == "_ACTIVE":
        provenance = window_state["active_provenance"]

    vol = get_channel_specific_slice_3D(dataset=dataset, t=window_state["t_idx"], channel=channel)
    center = (window_state["z"], window_state["y"], window_state["x"])
    radius = (r_z, r_y, r_x)
    blur_sigma = (blur_z, blur_y, blur_x)

    new_coords = get_nearby_max(vol, center, radius, blur_sigma) # type: ignore

    if new_worldline:
        worldline = worldlines.insert(Worldline())
        worldline_id = worldline.id

    else:
        worldline_id = window_state["selected_worldline"] or 0

    A = Annotation(
        t_idx=window_state["t_idx"],
        x=new_coords[2],
        y=new_coords[1],
        z=new_coords[0],
        worldline_id=worldline_id,
        parent_id=0,
        provenance=provenance
    )
    a = annotations.insert(A)

    return [
        {"type": "annotations/get_annotations"},
        {"type": "worldlines/get_worldlines"},
        {
            "type": "annotation_window/set_selected_annotation_local",
            "payload": int(a.id)
        }
    ]



from collections import OrderedDict
import torch.nn as nn
import torch.optim as optim
from tqdm.notebook import tqdm

from .build_pdists import get_all_pdists
from ..models.losses import *
from ..utils.utils import *


def get_optimization_trajectory(
    frame_to_optimize,
    parent,
    reference,
    container,
    results,
    zephir,
    zephod,
    clip_grad,
    lambda_t,
    lambda_d,
    lambda_n,
    lambda_n_mode,
    lr_ceiling,
    lr_coef,
    lr_floor,
    n_epoch,
    n_epoch_d,
    kernel_size=(3, 9, 9),
    lambda_n_decay=1.0,
    lr_step_size=10,
    lr_gamma=0.5,
    nb_delta=(2, 1),
    nb_epoch=5,
    sigmas=(1, 4, 4),):

    # pull variables from container
    dataset = container.get('dataset')
    allow_rotation = container.get('allow_rotation')
    annot = container.get('annot')
    channel = container.get('channel')
    covar = container.get('covar')
    dev = container.get('dev')
    gamma = container.get('gamma')
    img_shape = container.get('img_shape')
    neighbors = container.get('neighbors')
    shape_t = container.get('shape_t')
    t_annot = container.get('t_annot')
    z_compensator = container.get('z_compensator')

    trajectory = np.zeros((n_epoch+n_epoch_d, *results.shape[1:]))

    zephir.to(dev)

    t = frame_to_optimize
    root = np.where(t_annot == reference)[0][0]

    # calculate initial learning rate based on parent-child distance
    d_full = get_all_pdists(dataset, shape_t, channel, pbar=True)
    distance = d_full[t, parent]
    lr_init = min(max(lr_coef * distance, lr_floor), lr_ceiling)
    tqdm.write(f'\nFrame #{t}\t\tParent #{parent}'
               f'\t\tReference #{t_annot[root]}'
               f'\t\tDistance to parent: d={distance:.4f}')

    t_patch = [t]

    with torch.no_grad():

        # compiling Zephod results for feature detection loss, L_D
        input_tensor_d = []
        if lambda_d > 0 and zephod is not None:
            for _t in t_patch:
                data = get_data(dataset, _t, g=gamma, c=channel)
                _pred = zephod(data)
                _pred = (torch.max(_pred) - _pred) / torch.max(_pred)
                input_tensor_d.append(torch.where(_pred < 0.5, 0., 0.5))
            input_tensor_d = torch.stack(input_tensor_d, dim=0)
            input_tensor_d = blur3d(input_tensor_d, (1, 5, 5), (1, 2, 2), dev=dev)
            input_tensor_d.requires_grad = False

        # compiling target descriptors from reference frame for image registration, L_R
        data = get_data(dataset, t_annot[root], g=gamma, c=channel)
        vol = to_tensor(data, n_dim=5, dev=dev)
        zephir.theta.zero_()
        zephir.rho.zero_()
        zephir.rho[:len(t_patch)].add_(
            to_tensor(annot[root], dev=dev).expand(len(t_patch), -1, -1)
        )
        target_tensor = zephir(vol.expand(len(t_patch), -1, -1, -1, -1, -1))

        # loading input volumes
        input_tensor = []
        for _t in t_patch:
            data = get_data(dataset, _t, g=gamma, c=channel)
            vol = to_tensor(data, n_dim=5, grad=False, dev=dev)
            input_tensor.append(vol)
        input_tensor = torch.stack(input_tensor, dim=0)

        # initializing model parameters at parent coordinates
        xyz_parent = to_tensor(results[parent, ...], dev=dev)
        zephir.theta.zero_()
        zephir.rho.zero_()
        zephir.rho[:len(t_patch)].add_(
            xyz_parent.expand(len(t_patch), -1, -1)
        )

        # compiling spring connections and stiffnesses for spatial regularization, L_N
        ind = neighbors[root][:, :, 0]
        d_ref = to_tensor(neighbors[root][:, :, 1:], dev=dev)
        if covar is not None:
            nn_covar = np.empty_like(ind)
            for k in range(ind.shape[1]):
                nn_covar[:, k] = covar[range(ind.shape[0]), ind[:, k].astype(int)]/20
            k_ij = to_tensor(
                nn_covar /
                np.clip(np.max(nn_covar, axis=-1)[:, None], 1, None),
                grad=False, dev=dev
            )
            k_ij = lambda_n * torch.relu(k_ij)
        else:
            k_ij = lambda_n * torch.ones_like(
                to_tensor(ind), requires_grad=False, device=dev
            )

        # resetting optimizer state
        optimizer = optim.SGD(zephir.parameters(), lr=lr_init)
        scheduler = optim.lr_scheduler.StepLR(optimizer, lr_step_size, lr_gamma)
        _kernel_size = kernel_size
        _sigmas = sigmas

    # training loop here
    zephir.train()
    pbar = tqdm(range(n_epoch + n_epoch_d), desc='Tracking', unit='epochs')
    for epoch in pbar:

        if (epoch + 1) % nb_epoch == 0:
            with torch.no_grad():
                # updating spring stiffness
                if lambda_n_decay < 1.0:
                    k_ij *= lambda_n_decay
                # updating kernels for dynamic Gaussian blurring
                _kernel_size = (3, _kernel_size[1] - nb_delta[0], _kernel_size[2] - nb_delta[0])
                _sigmas = (1, _sigmas[1] - nb_delta[1], _sigmas[1] - nb_delta[1])
                if _kernel_size[1] <= 1 or _sigmas[1] <= 1:
                    # resets kernels to be inactive when it reaches a threshold
                    _kernel_size = (1, 1, 1)
                    _sigmas = (1, 1, 1)

        optimizer.zero_grad()

        # using automatic matrix precision casting to dynamically switch
        # between float16 and float32 to minimize memory consumption
        with torch.autocast(device_type=dev):

            # sampling child descriptors at model parameters from child frames
            pred = zephir(input_tensor)

            # dynamic Gaussian blurring
            blur_target = blur3d(target_tensor, _kernel_size, _sigmas, dev=dev)
            blur_pred = blur3d(pred, _kernel_size, _sigmas, dev=dev)

            if epoch > n_epoch and lambda_d > 0 and input_tensor_d is not None:
                # feature detection loss, L_D
                loss = torch.mean(reg_d(
                    lambda_d, input_tensor_d,
                    zephir.rho, zephir.n_chunks
                ))
            else:
                # image registration loss, L_R
                loss = torch.mean(corr_loss(blur_pred, blur_target))

                reg = 0
                if torch.max(k_ij) > 0:
                    # spatial regularization, L_N
                    reg += torch.mean(
                        reg_n(k_ij, zephir.rho, None, ind, d_ref,
                              None, lambda_n_mode)
                    )
                if lambda_t > 0:
                    # temporal smoothing, L_T
                    reg += torch.mean(reg_t(lambda_t, blur_pred, (1, 5, 5), None))
                loss += reg

        # backpropagation call
        loss.backward()
        if clip_grad > 0.:
            nn.utils.clip_grad_value_(zephir.parameters(), clip_grad)
        if z_compensator > 0.:
            # due to anisotropy in depth-axis commonly found in 3D datasets,
            # gradient descent may not properly update parameters in that axis
            # so we multiply calculated gradients in that axis by (1+z_compensator)
            # to compensate
            with torch.no_grad():
                zephir.rho[:, :, -1] += (
                    - z_compensator
                    * optimizer.param_groups[0]['lr']
                    * zephir.rho.grad[:, :, -1]
                )

        # gradient descent step, update model parameters
        optimizer.step()
        scheduler.step()

        with torch.no_grad():
            current_lr = optimizer.param_groups[0]['lr']
            pbar.set_postfix(OrderedDict(
                Loss=f'{loss.item():.4f}',
                Loss_R=f'{(loss - reg).item():.4f}',
                LR=f'{current_lr:.4f}'
            ))

            # resetting model parameters for rotation
            if not allow_rotation:
                zephir.theta.zero_()

            for n in range(zephir.rho.shape[1]):
                x, y, z = get_pixel(zephir.rho.clone()[0, n, :], img_shape)
                trajectory[epoch, n, :] = np.array([x, y, z])

    return trajectory



"""
extract_traces.py: Extract fluorescent activity traces.

Using keypoint coordinates in annotations.h5, both manual annotations and Zephir
results, extracts fluorescent activity traces by analyzing pixel intensities.
This uses build arguments and container attributes from checkpoint.pt.
A modified version of the Zephir model is used to grid sample pixels around each
annotation and exclude pixels from neighboring cells.

Adapted from program by Mahdi Torkashvand

Usage:
    extract_traces -h | --help
    extract_traces -v | --version
    extract_traces --dataset=<dataset> [options]

Options:
    -h --help                           show this message and exit.
    -v --version                        show version information and exit.
    --dataset=<dataset>                 path to data directory to analyze.
    --channel=<channel>                 data channel to pull traces from.
    --cuda=<cuda>  						check if a CUDA-compatible GPU is available for  use. [default: True]
    --cutoff=<cutoff>  					cutoff frequency for lowpass filter. [default: 1.0]
    --debleach=<debleach>  				whether to apply a debleaching step. [default: True]
    --dist_thresh=<dist_thresh>  		threshold for pw-frame distance to exclude. [default: 2.1]
    --n_chunks=<n_chunks>               number of steps to divide the forward pass into. [default: 10]
    --n_cluster=<n_cluster>     		threshold for clustering for heatmap. [default: 1]
    --nn_max=<nn_max>  					maximum number of neighbors to mask. [default: 5]
    --rma_channel=<rma_channel>  		data channel for removing multiplicative artifacts.
    --t_list=<t_list>  					time points to analyze.
    --wlid_ref=<wlid_ref>               subset of neurons to extract by worldline_id.
"""

from docopt import docopt
from scipy.optimize import curve_fit
from scipy.signal import butter, filtfilt
from skimage.transform import resize
from sklearn.cluster import SpectralClustering
import sklearn.neighbors as sk

from ..__version__ import __version__
from . import *
from ..models.zephir import ZephIR
from ..utils.utils import *


def double_exp(t, a1, b1, a2, b2):
    return a1 * np.exp(-b1 * t) + a2 * np.exp(-b2 * t)


def extract_traces(
    dataset,
    cuda,
    channel,
    cutoff,
    debleach,
    n_chunks,
    n_cluster,
    nn_max,
    rma_channel,
    t_list,
    dist_thresh,
    wlid_ref,
    a_scalar=0.8,
    crop_shape=(3, 9, 9),
    eps=0.01,
    npx_to_keep=16,
    save_as_npy=True,
    save_as_fig=True,
    verbose=False):

    # checking for available CUDA GPU
    if cuda in ['True', 'Y', 'y'] and torch.cuda.is_available():
        print('\n*** GPU available!')
        dev = 'cuda'
    else:
        dev = 'cpu'
    print(f'\nUsing device: {dev}\n\n')

    container = get_checkpoint(dataset, 'container', verbose=True)
    container.update({
        'dev': dev,
        'exclude_self': False,
        'exclusive_prov': None,
    })
    container, results = build_annotations(
        container=container,
        annotation=None,
        t_ref=None,
        wlid_ref=wlid_ref,
        n_ref=None,
    )

    s_list = container.get('s_list')
    worldline_id = container.get('worldline_id')

    if t_list is None:
        t_list = list(container.get('t_annot'))

    # import data here
    print('Loading file...')
    metadata = get_metadata(dataset)
    img_shape = (metadata['shape_z'], metadata['shape_y'], metadata['shape_x'])
    shape_t = metadata['shape_t']
    shape_c = metadata['shape_c']
    n_neuron = results.shape[1]

    crop_spacing = tuple(np.array(crop_shape) / np.array(img_shape))
    nn_max = min(nn_max, n_neuron - 1)

    print(f'\nCompiling model...')
    model_kwargs = {
        'allow_rotation': False,
        'dimmer_ratio': 1.0,
        'fovea_sigma': (-1, -1, -1),
        'grid_shape': crop_shape,
        'grid_spacing': crop_spacing,
        'n_chunks': n_chunks,
        'n_frame': 1,
        'shape_n': n_neuron,
        'ftr_ratio': 1.0,
        'ret_stride': 1.0,
    }
    model = ZephIR(**model_kwargs).to(dev)

    if dist_thresh < 2.0:
        t_ignore = np.where(s_list > dist_thresh)[0]
        print(f'*** Excluding traces for frames: {list(t_ignore)}')
        t_list = np.setdiff1d(t_list, t_ignore)

    traces = np.empty((shape_c, n_neuron, shape_t)) * np.nan
    for t in tqdm(t_list, desc='Compiling raw traces', unit='frames'):
        data = get_slice(dataset, t).astype(float)
        with torch.no_grad():
            vol = to_tensor(data, n_dim=5, grad=False, dev=dev)
            model.theta.zero_()
            model.rho.zero_()
            model.rho.add_(to_tensor(results[t, ...], dev=dev).unsqueeze(0))
            descriptors = model(vol.unsqueeze(0))[0]
            descriptors = descriptors.view(n_neuron, shape_c, -1)
            grids = (model.grid + model.rho.view((-1, 1, 1, 1, 3))).view(n_neuron, -1, 3)
            grids *= to_tensor(np.array(img_shape)[::-1] / np.max(img_shape), n_dim=3, dev=dev)

            tree = sk.KDTree(results[t, ...])
            ind = (tree.query(results[t, ...], k=nn_max + 1, return_distance=False))[:, 1:]
            for n in range(n_neuron):
                mask = descriptors[n, ...].int()
                if nn_max > 0:
                    nn_idx = ind[n, :]
                    cdists = torch.cdist(
                        grids[n].unsqueeze(0).expand((nn_max, -1, -1)),
                        torch.stack([grids[n_] for n_ in nn_idx], dim=0)
                    )
                    for d in cdists:
                        mask[:, torch.mean(d, dim=-1) < eps] *= 0
                for c in range(shape_c):
                    nz = to_numpy(mask[c][torch.nonzero(mask[c], as_tuple=True)]).flatten()
                    if len(nz) == 0:
                        traces[c, n, t] = np.nan
                    else:
                        traces[c, n, t] = np.mean(np.sort(nz)[-npx_to_keep:]) / 255
    if save_as_npy:
        np.save(str(dataset / f'traces_raw.npy'), traces)

    for n in range(n_neuron):
        for c in range(shape_c):
            traces[c, n, t_list] = fill_nans(traces[c, n, t_list])

    # remove multiplicative artifact
    if rma_channel is not None:
        for n in tqdm(range(n_neuron), desc='Removing artifacts', unit='items'):
            params_r = curve_fit(
                double_exp, np.arange(len(t_list)),
                traces[rma_channel, n, t_list],
                bounds=(0., np.inf), method='trf'
            )[0]
            curve_r = double_exp(np.arange(len(t_list)), *params_r)
            if verbose:
                # plt.plot(t_list, traces[0, n, t_list], 'r-')
                plt.plot(t_list, traces[1, n, t_list], 'g-')
                # plt.plot(t_list, curve_r, 'r-', alpha=0.8)
                plt.show()

            artifacts = (traces[rma_channel, n, t_list] - curve_r) / curve_r
            traces[rma_channel, n, t_list] /= 1 + artifacts
            if channel is not None:
                traces[channel, n, t_list] /= 1 + a_scalar * artifacts
            else:
                for c in range(traces.shape[0]):
                    if c != rma_channel:
                        traces[c, n, t_list] /= 1 + a_scalar * artifacts
            for c in range(traces.shape[0]):
                traces[c, n, t_list] = fill_nans(traces[c, n, t_list])

    if channel is not None:
        traces = traces[None, channel, ...]

    # lowpass filter
    if cutoff > 0:
        for n in tqdm(range(n_neuron), desc='Applying lowpass filter', unit='items'):
            params = butter(N=2, Wn=cutoff, output='ba', btype='lowpass', fs=4.0, analog=False)
            for c in range(traces.shape[0]):
                traces[c, n, t_list] = filtfilt(
                    params[0], params[1], traces[c, n, t_list]
                )
                traces[c, n, t_list] = fill_nans(traces[c, n, t_list])

            if verbose:
                plt.plot(t_list, traces[1, n, t_list], 'g-')
                plt.show()

    # debleaching
    if debleach:
        for n in tqdm(range(n_neuron), desc='Debleaching', unit='items'):
            for c in range(traces.shape[0]):
                params = curve_fit(
                    double_exp,
                    np.arange(len(t_list)),
                    traces[c, n, t_list],
                    bounds=(0., np.inf),
                    method='trf'
                )[0]
                curve = double_exp(np.arange(len(t_list)), *params)
                curve = np.clip(curve, np.nanmin(traces[c, n, t_list]) + 0.002, 1.0)
                if verbose:
                    plt.plot(t_list, traces[c, n, t_list], 'g-')
                    plt.plot(t_list, curve, 'g-', alpha=0.8)
                    plt.show()
                traces[c, n, t_list] *= curve[0, None] / curve
                traces[c, n, t_list] = fill_nans(traces[c, n, t_list])

    traces = (traces - traces[:, :, t_list[0], None]) / traces[:, :, t_list[0], None]
    for n in range(n_neuron):
        traces[0, n, :] = fill_nans(traces[0, n, :])

    if save_as_npy:
        np.save(str(dataset / f'traces.npy'), traces)

    if save_as_fig:
        # plt.rcParams.update({'font.size': 14})
        plt.figure(figsize=(5, n_neuron))
        colors = plt.cm.hsv(np.linspace(0, 1, n_neuron + 1)[:-1])[:, :3]
        offset = np.cumsum(np.append(0.0,
                                     np.nanmax(traces[0, :-1, :], axis=-1)
                                     - np.nanmin(traces[0, :-1, :], axis=-1)
                                     ))
        s = traces[0] - np.nanmin(traces[0], axis=-1)[:, None] + offset[:, None]
        for n in range(n_neuron):
            plt.plot(s[n], linewidth=1, color=colors[n])
        plt.yticks(s[:, 0], [str(n) for n in worldline_id])
        plt.ylim(np.nanmin(s) - 0.01, np.nanmax(s) + 0.01)
        plt.xlabel('Time (frames)')
        plt.ylabel('Neuron ID')
        plt.savefig(str(dataset / f'traces.png'), format='png')
        plt.close('all')

        plt.figure(figsize=(10, 5))
        s = (traces[0] - np.nanmin(traces[0], axis=-1)[:, None])
        s = s / np.clip(np.nanmax(s, axis=-1)[:, None], 3.0, 12.0)
        for n in range(n_neuron):
            s[n, :] = fill_nans(s[n, :])
        print(np.mean(s), np.min(s), np.max(s))
        sm = np.zeros((n_neuron, n_neuron))
        for i in range(n_neuron):
            for j in range(i, n_neuron):
                sm[i, j] = np.corrcoef(s[i, :], s[j, :])[0, 1] + 1
        sm = sm + np.transpose(sm)
        sc = SpectralClustering(n_cluster, affinity='precomputed')
        fc = sc.fit_predict(sm)
        hm = s[np.argsort(fc), :]
        # plt.imshow(hm)
        plt.imshow(resize(hm, (hm.shape[0], 2 * hm.shape[0])))
        plt.yticks([])
        plt.xlabel('Time (frames)')
        plt.ylabel('Neuron ID')
        plt.tight_layout()
        plt.savefig(str(dataset / f'traces_hm.png'), format='png', dpi=500)
        plt.close('all')

    return traces


def main():
    args = docopt(__doc__, version=f'Zephir extract_traces {__version__}')
    print(args, '\n')

    extract_traces(
        dataset=Path(args['--dataset']),
        channel=int(args['--channel']) if args['--channel'] else None,
        cuda=args['--cuda'] in ['True', 'Y', 'y'],
        cutoff=float(args['--cutoff']),
        debleach=args['--debleach'] in ['True', 'Y', 'y'],
        n_chunks=int(args['--n_chunks']),
        n_cluster=int(args['--n_cluster']),
        nn_max=int(args['--nn_max']),
        rma_channel=int(args['--rma_channel']) if args['--rma_channel'] else None,
        t_list=list(args['--t_list']) if args['--t_list'] else None,
        dist_thresh=float(args['--dist_thresh']),
        wlid_ref=eval(args['--wlid_ref']) if args['--wlid_ref'] else None,
    )


if __name__ == '__main__':
    main()



# These are tools to be used along with Python dataclasses.
#
# Author: vivekv2@gmail.com

from dataclasses import asdict
from datetime import datetime
from pathlib import Path
from typing import Type, TypeVar

import pandas as pd
import numpy as np
import h5py


def add_backup_timestamp(path: Path) -> Path:
    timestamp = datetime.now().strftime("_%Y_%m_%d_%H_%M_%S")
    backup_folder = path.parent / "backup"
    backup_folder.mkdir(exist_ok=True)
    return backup_folder / (path.stem + timestamp + path.suffix)


def archive_file(path: Path):
    if path.exists():
        path.rename(add_backup_timestamp(path))


DT = TypeVar('DT', bound='DataclassTableBase')  # For classmethod typing.


class DataclassTableBase:
    """This class needs to be subclassed, and a valid dataclass has to be
    assigned to row_class."""

    row_class: type

    def __init__(self, data: pd.DataFrame = None):

        self.column_types = dict(self.row_class.__annotations__)

        self.df = pd.DataFrame(columns=self.column_types)

        if data is not None:
            self.df = pd.concat([self.df, data])

        self.next_id = np.max(self.df['id']) + 1
        if np.isnan(self.next_id):
            self.next_id = 1

        self._fix_types()

    def _fix_types(self):
        """The index type may be coerced by pandas (e.g. u64 instead of u32),
        so that needs to be set after resetting it."""

        for k in self.df.columns:
            self.df[k] = self.df[k].astype(self.column_types[k])

    def _insert_and_preserve_id(self, data):

        row = pd.DataFrame(asdict(data), index=(0, ))
        self.df = pd.concat([self.df, row], ignore_index=True)
        return data

    def insert(self, data):

        data.id = np.uint32(self.next_id)
        row = pd.DataFrame(asdict(data), index=(0, ))
        self.df = pd.concat([self.df, row], ignore_index=True)
        self.next_id = self.next_id + 1
        return data

    def object_from_row(self, row: pd.DataFrame):
        row_dict = {k: list(v.items())[0][1] for k, v in row.to_dict().items()}
        return self.row_class(**row_dict)

    def get_row(self, idx: int):

        row = self.df.loc[(idx, ), :]
        return self.object_from_row(row)

    def get_first(self):

        row = self.df.iloc[[0,], :]
        return self.object_from_row(row)

    def get_row_idx(self, id: int):

        row = self.filter(lambda x: x["id"] == id).df
        return row.index[0]

    def get(self, id: int):

        row = self.filter(lambda x: x["id"] == id)
        return self.object_from_row(row.df)

    def update(self, id: int, data: dict):

        row_idx = self.get_row_idx(id)

        new_data = asdict(self.get(id))
        new_data.update(data)
        new_row = pd.DataFrame(new_data, index=(row_idx, ))
        self.df.update(new_row)

        return self.get(id)

    def delete(self, id: int):

        row_idx = self.get_row_idx(id)
        self.df = self.df.drop(row_idx)

        return id

    def delete_ids(self, ids: list):

        rows_to_delete = self.df[self.df["id"].apply(lambda x: x in ids)].index
        self.df = self.df.drop(rows_to_delete)

    def filter(self: DT, fn) -> DT:

        df = self.df[fn(self.df)]
        return type(self)(df)

    def to_hdf(self, filename: Path, archive=True):

        filename = Path(filename)

        if archive:
            archive_file(filename)

        f = h5py.File(filename, "w")
        N = self.df.shape[0]

        for c in self.df.columns:

            if self.column_types[c] == np.string_:
                dtype = "S8"
            else:
                dtype = self.column_types[c]

            f.create_dataset(c,
                             shape=(N, ),
                             dtype=dtype,
                             data=self.df[c].astype(dtype))

    def __len__(self):
        return len(self.df)

    def __repr__(self):
        return repr(self.df)

    def _repr_html_(self):
        return self.df._repr_html_()

    def __iter__(self):
        self._iter_idx = 0
        return self

    def __next__(self):
        if self._iter_idx < len(self.df):
            val = self.get_row(self.df.index[self._iter_idx])
            self._iter_idx = self._iter_idx + 1
            return val
        else:
            raise StopIteration

    @classmethod
    def from_hdf(cls: Type[DT], filename: Path) -> DT:

        filename = Path(filename)

        f = h5py.File(filename, "r")

        data = pd.DataFrame()
        for k in f:
            data[k] = f[k]

        return cls(data)



"""
auto_annotate.py: use ZephOD's feature detection to automatically annotate a frame in a dataset.

Usage:
    auto_annotate -h | --help
    auto_annotate -v | --version
    auto_annotate --dataset=<dataset> [options]

Options:
    -h --help                           show this message and exit.
    -v --version                        show version information and exit.
    --dataset=<dataset>                 path to data directory to analyze.
    --tidx=<tidx>                       frame number or time index to analyze. [default: 0]
    --channel=<channel>                 data channel to register.
    --cuda=<cuda>                       check if a CUDA-compatible GPU is available for  use. [default: True]
    --min_distance=<min_distance>       minimum distance between detected features in pixels. [default: 4]
    --min_val=<min_val>                 minimum intensity value for a peak. [default: 1]
    --model=<model>                     path to checkpoint for model.
"""


from docopt import docopt
import shutil
from skimage.feature import peak_local_max

from ..__version__ import __version__
from ..utils.utils import *
from ..zephod.model import ZephOD


def auto_annotate(dataset: Path, tidx: int, args: dict):

    if torch.cuda.is_available() and args['--cuda'] in ['True', 'Y', 'y']:
        # Moving to GPU
        print('\n*** GPU available!')
        dev = torch.device('cuda:0')
    else:
        dev = torch.device('cpu')

    if args['--model'] is None:
        checkpoint_path = Path(__file__).parent.parent / 'zephod' / 'model.pt'
    else:
        checkpoint_path = Path(args['--model'])
    if Path.is_file(checkpoint_path):
        print('\nPrevious checkpoint available.')
        try:
            checkpoint = torch.load(checkpoint_path)
        except RuntimeError:
            print('*** DEVICE INCOMPATIBLE! Mapping CUDA tensors to CPU...')
            checkpoint = torch.load(checkpoint_path, map_location=torch.device('cpu'))
    else:
        print('\n******* ERROR: model not found! Exiting...')
        exit()

    # loading other user arguments
    channel = int(args['--channel']) if args['--channel'] else None
    min_distance = int(args['--min_distance'])

    # performing inference with ZephOD
    # this produces a probability map of the desired features
    print('\nRunning feature detection with ZephOD...')
    model = ZephOD(**checkpoint['model_kwargs'])
    model.load_state_dict(checkpoint['state_dict'])
    model.to(dev)
    model.eval()
    vol = get_data(dataset, t=tidx, c=channel)
    pred = model(vol)
    im = to_numpy(pred[0, 0] * 255)
    peaks = peak_local_max(im, min_distance=min_distance)

    # removing peaks at which the pixel intensity does not meet a minimum
    low_peak_list = []
    for i, (z, y, x) in enumerate(peaks):
        if np.max(vol[:, z, y, x]) < float(args['--min_val'])/255:
            low_peak_list.append(i)
    peaks = np.delete(peaks, low_peak_list, axis=0)

    print(f'*** Found {peaks.shape[0]} keypoints!')

    # compiling results into annotator-ready format
    shape_n = peaks.shape[0]
    img_shape = np.array(im.shape)
    provenance = np.array([b'ZEIR'] * shape_n)
    xyz_pd = np.concatenate(
        (tidx * np.ones((shape_n, 1)),
         peaks[:, ::-1] / img_shape[None, ::-1],
         np.arange(shape_n)[:, None],
         -1 * np.ones((shape_n, 1)),
         provenance[:, None]),
        axis=-1
    )

    # creating backup
    if not (dataset / 'backup').is_dir():
        Path.mkdir(dataset / 'backup')
    if (dataset / 'annotations.h5').is_file():
        print('\n\n*** WARNING: Existing annotations file found!\n'
              'This process will create a backup of the existing file and '
              'overwrite any annotations in the selected frame with new auto-detected annotations.\n'
              'Please note that the new annotations are NOT properly linked or '
              'assigned to any existing worldlines.')
        now = datetime.datetime.now()
        now_ = now.strftime("%m_%d_%Y_%H_%M_%S")
        shutil.copy(dataset / 'annotations.h5',
                    dataset / 'backup' / f'annotations_{now_}.h5')

        # adding any existing annotations to the new annotation set
        # this skips the selected frame, essentially
        annotation = get_annotation_df(dataset)
        for t in np.unique(annotation['t_idx']):
            if t == tidx:
                continue
            u, annot, prov = get_annotation(annotation, t, None, False)
            xyz_pd = np.append(
                xyz_pd,
                np.concatenate(
                    (np.ones((len(u), 1)) * t,
                     annot / 2 + 0.5,
                     u[:, None],
                     -1 * np.ones((len(u), 1)),
                     prov[:, None]),
                    axis=-1
                ),
                axis=0
            )

    # writing results to file
    columns = {
        't_idx': np.uint32,
        'x': np.float32,
        'y': np.float32,
        'z': np.float32,
        'worldline_id': np.uint32,
        'parent_id': np.uint32,
        'provenance': np.dtype("S4"),
    }
    f = h5py.File(dataset / 'annotations.h5', mode='w')
    data = np.array(list(range(1, xyz_pd.shape[0] + 1)), dtype=np.uint32)
    f.create_dataset('id', shape=(xyz_pd.shape[0],), dtype=np.uint32, data=data)
    for i, c in enumerate(columns.keys()):
        if c == 'provenance':
            data = np.array(xyz_pd[:, i], dtype=columns[c])
        else:
            data = np.array(xyz_pd[:, i].astype(np.float32), dtype=columns[c])
        f.create_dataset(c, shape=(xyz_pd.shape[0],), dtype=columns[c], data=data)
    f.close()

    print('\n\n*** DONE!')
    return


def main():
    args = docopt(__doc__, version=f'ZephIR auto_annotate {__version__}')
    # print(args, '\n')

    dataset = Path(args['--dataset'])

    auto_annotate(
        dataset=dataset,
        tidx=int(args['--tidx']),
        args=args
    )


if __name__ == '__main__':
    main()



from logging import Logger
from pathlib import Path
from typing import List
import re

import numpy as np  # needed in eval

from ..data import (AnnotationTable, WorldlineTable)
from ._utilities import default_args

@default_args("_this, _now, *, _ACTIVE")
def change_provenance(dataset: Path, annotations: AnnotationTable,
                      worldlines: WorldlineTable, window_state, arg: str,
                      logger: Logger) -> List[dict]:
    """
    arg: worldline_id, times, old_provenance, new_provenance
        worldline_id: worldline_id to delete.
            '*': all worldlines
            '_this': currently active worldline [default]
        times: times to delete.
            '*': all times
            '3:5': times 3 and 4
            '3:_now': times 3 through current time (minus 1)
        old_provenance: Any regexp for matching.
            _ACTIVE: Use the currently active provenance [default]
            '*': all provenances
            'XX..' Match XX followed by any two characters.
        new_provenance: Any regexp for matching.
            _ACTIVE: Use the currently active provenance [default]
            'XX..' Match XX followed by any two characters.
    ex: "*, _now, NEIR, ANTT": Updates all neurons in this frame with provenance "NEIR" to "ANTT"
    """

    arg_list = arg.replace(" ", "").split(",")
    bare_worldline = arg_list[0]
    bare_times = arg_list[1].replace("_now", str(window_state["t_idx"]))
    bare_old_provenance = arg_list[2].replace("_ACTIVE",
                                              window_state["active_provenance"])
    bare_new_provenance = arg_list[3].replace("_ACTIVE",
                                              window_state["active_provenance"])

    if bare_worldline == "*":
        bad_wl = True
    elif bare_worldline == "_this":
        bad_wl = annotations.df["worldline_id"] == window_state[
            "selected_worldline"]
    else:
        bad_wl = annotations.df["worldline_id"] == int(bare_worldline)

    if bare_times == "*":
        bad_times = True
    else:
        all_times = range(window_state["shape_t"])
        slice_ = eval(f"np.s_[{bare_times}]")
        times = all_times[slice_]
        if type(times) is int:
            bad_times = annotations.df["t_idx"] == times
        else:
            bad_times = annotations.df["t_idx"].apply(lambda x: x in times)

    if bare_old_provenance == "*":
        bad_prov = True
    else:
        pattern = re.compile(bare_old_provenance)
        bad_prov = annotations.df["provenance"].apply(
            lambda x: bool(pattern.match(x.decode())))

    to_change = annotations.df[bad_wl & bad_times & bad_prov].index

    annotations.df.provenance[to_change] = bare_new_provenance.encode('utf-8')

    return [{"type": "annotations/get_annotations"}]


"""
overwrite_checkpoint.py: overwrite existing key and value in ZephIR checkpoint.

Usage:
	overwrite_checkpoint.py -h | --help
	overwrite_checkpoint.py -v | --version
	overwrite_checkpoint.py --dataset=<dataset> --key=<key> --value=<value> [options]

Options:
	-h --help                           	show this message and exit.
	-v --version                        	show version information and exit.
	--dataset=<dataset>  					path to data directory to analyze.
	--key=<key>  							name of item in checkpoint to overwrite.
	--value=<value>  						new item to write into checkpoint.
"""

from docopt import docopt

from ..__version__ import __version__

from . import *
from ..models.container import Container
from ..utils.utils import *


def overwrite_checkpoint(dataset, key, value):
	_checkpoint = load_checkpoint(dataset, fallback=False, verbose=True)

	updated_items = 0

	if key == 'args' and value in ['all', '', None]:
		if (dataset / 'args.json').is_file():
			with open(str(dataset / 'args.json')) as json_file:
				args = json.load(json_file)
		update_checkpoint(dataset, {'args': args})
		updated_items += 1

	if key in _checkpoint:
		print('Updating checkpoint main dict...')
		update_checkpoint(dataset, {key: type(_checkpoint[key])(value)})
		updated_items += 1

	if key in _checkpoint['args'] or f'--{key}' in _checkpoint['args']:
		print('Updating saved args...')
		args = _checkpoint['args']
		args[f'--{key}'] = type(args[f'--{key}'])(value)
		update_checkpoint(dataset, {'args': args})
		updated_items += 1

		_args_list = [
			'allow_rotation', 'channel', 'exclude_self', 'exclusive_prov',
			'gamma', 'include_all', 'n_frame', 'z_compensator'
		]
		if key in _args_list:
			print('Rebuilding container from updated args...')
			container = Container(
				dataset=dataset,
				allow_rotation=args['--allow_rotation'] in ['True', 'Y', 'y'],
				channel=int(args['--channel']) if args['--channel'] else None,
				dev='cpu',
				exclude_self=args['--exclude_self'] in ['True', 'Y', 'y'],
				exclusive_prov=(bytes(args['--exclusive_prov'], 'utf-8')
								if args['--exclusive_prov'] else None),
				gamma=float(args['--gamma']),
				include_all=args['--include_all'] in ['True', 'Y', 'y'],
				n_frame=int(args['--n_frame']),
				z_compensator=float(args['--z_compensator']),
			)
			updated_items += 1
		else:
			container = _checkpoint['container']
		_args_list = _args_list + ['t_ref', 'wlid_ref', 'n_ref']
		if key in _args_list:
			build_annotations(
				container=container,
				annotation=None,
				t_ref=eval(args['--t_ref']) if args['--t_ref'] else None,
				wlid_ref=eval(args['--wlid_ref']) if args['--wlid_ref'] else None,
				n_ref=int(args['--n_ref']) if args['--n_ref'] else None,
			)
			updated_items += 1
		_args_list = _args_list + ['dimmer_ratio', 'grid_shape', 'fovea_sigma', 'n_chunks']
		if key in _args_list:
			build_models(
				container=container,
				dimmer_ratio=float(args['--dimmer_ratio']),
				grid_shape=(5, 2 * (int(args['--grid_shape']) // 2) + 1,
							2 * (int(args['--grid_shape']) // 2) + 1),
				fovea_sigma=(1, float(args['--fovea_sigma']),
							 float(args['--fovea_sigma'])),
				n_chunks=int(args['--n_chunks']),
			)
			updated_items += 1
		_args_list = _args_list + ['load_nn', 'nn_max']
		if key in _args_list:
			build_springs(
				container=container,
				load_nn=args['--load_nn'] in ['True', 'Y', 'y'],
				nn_max=int(args['--nn_max']),
			)
			updated_items += 1
		_args_list = _args_list + ['sort_mode', 't_ignore']
		if key in _args_list:
			build_tree(
				container=container,
				sort_mode=str(args['--sort_mode']),
				t_ignore=eval(args['--t_ignore']) if args['--t_ignore'] else None,
			)
			updated_items += 1

	if updated_items == 0:
		print('*** KEY NOT FOUND! Check arguments and try again.')

	return


def main():
	args = docopt(__doc__, version=f'ZephIR overwrite_checkpoint {__version__}')
	print(args, '\n')

	overwrite_checkpoint(
		dataset=Path(args['--dataset']),
		key=str(args['--key']),
		value=eval(args['--value']),
	)


if __name__ == '__main__':
	main()



# Utilities to visualize images from ND arrays.
#
# Authors: vivekv2@gmail.com, maedeh.seyedolmohadesin@gmail.com

from math import floor
from typing import Tuple, List, Union

import numpy as np
from scipy.ndimage import rotate
from scipy.ndimage.filters import convolve

from .shapes import gaussian

def blur(vol:np.ndarray, sigma:np.ndarray) -> np.ndarray:
    """Blur the given volume using a gaussian kernal with specified width. The
    shape of vol is preserved."""

    sigma = np.array(sigma)

    g = gaussian(sigma, norm="area")

    return convolve(vol, g, mode="constant", cval=0.0)

def apply_lut(x: np.ndarray, lo: float, hi: float, newtype=None) -> np.ndarray:
    """Clip x to the range [lo, hi], then rescale to fill the range of
    newtype."""

    if newtype is None:
        newtype = x.dtype

    y_float = (x-lo)/(hi-lo)
    y_clipped = np.clip(y_float, 0, 1)

    if np.issubdtype(newtype, np.integer):
        maxval = np.iinfo(newtype).max
    else:
        maxval = 1.0

    return (maxval*y_clipped).astype(newtype)

def auto_lut(x: np.ndarray, quantiles=(0.5,0.99), newtype=None) -> np.ndarray:
    """Linearly map the specified quantiles of x to the range of newtype."""

    lo = np.quantile(x, quantiles[0])
    hi = np.quantile(x, quantiles[1])

    return apply_lut(x, lo, hi, newtype=newtype)

def mip_x(vol:np.ndarray) -> np.ndarray:
    return np.transpose(np.max(vol, axis=2),
        (1, 0, *(range(2, np.ndim(vol)-1))))

def mip_y(vol:np.ndarray) -> np.ndarray:
    return np.max(vol, axis=1)

def mip_z(vol:np.ndarray) -> np.ndarray:
    return np.max(vol, axis=0)

def mip_threeview(vol: np.ndarray, scale=(4,1,1)) -> np.ndarray:
    """Combine 3 maximum intensity projections of a volume into a single
    2D array."""

    S = vol.shape[:3] * np.array(scale)
    output_shape = (S[1] + S[0],
                    S[2] + S[0])

    if vol.ndim == 4:
        output_shape = (*output_shape, 3)

    vol = np.repeat(vol, scale[0], axis=0)
    vol = np.repeat(vol, scale[1], axis=1)
    vol = np.repeat(vol, scale[2], axis=2)

    x = mip_x(vol)
    y = mip_y(vol)
    z = mip_z(vol)

    output = np.zeros(output_shape, dtype=vol.dtype)

    output[:S[1], :S[2]] = z
    output[:S[1], S[2]:] = x
    output[S[1]:, :S[2]] = y

    return output

def rgb_from_bw(x: np.ndarray, rgb=(1,1,1)) -> np.ndarray:
    """Convert a grayscale image to RGB."""

    return np.stack((rgb[0] * x, rgb[1] * x, rgb[2] * x), x.ndim)

def compare_red_green(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """Autoscale 2D arrays x and y, then stack them with a blank blue channel
    to form a 3D array with shape (3, size_Y, size_X) and type uint8."""

    r = auto_lut(x, newtype=np.uint8)
    g = auto_lut(y, newtype=np.uint8)
    b = np.zeros_like(r)

    return np.dstack([r,g,b])

def shift(x: np.ndarray, xyz=(0,0,0), xyz_axis=(2,1,0)) -> np.ndarray:
    """Shift x in x, y and z directions (wraps numpy roll)."""

    for i in range(3):
        x = np.roll(x, xyz[i], xyz_axis[i])

    return x

def rotate_z(vol:np.ndarray, angle: float, reshape=False) -> np.ndarray:
    """Rotate the given volume about z by angle (in degrees)."""

    return rotate(vol, angle, axes=(2, 1), reshape=reshape)

def rotate_180(vol:np.ndarray, axis: int, reshape=False) -> np.ndarray:
    """Rotates the given volume about the given axis by 180 degrees by
    flipping twice."""

    if axis == 1:
        axes = (0, 2)
    elif axis == 0:
        axes = (1, 2)
    else:
        axes = (0, 1)

    return rotate(vol, 180, axes=axes, reshape=reshape)

def _coord_from_idx(idx: int, shape: int) -> float:
    return (idx + 0.5)/shape

def _idx_from_coord(coord: float, shape: int) -> int:
    return max(floor(coord*shape - 1e-6), 0)

def coords_from_idx(idx: tuple, shape: tuple) -> tuple:
    """Return image coordinates in (0, 1) for the center of a pixel located
    at index pixel_idx in [0, shape-1]. The returned coordinates cannot be 0 or
    1 because the pixel centers cannot be located at the edges of the
    volume."""
    return tuple((_coord_from_idx(i, s) for (i, s) in zip(idx, shape)))

def idx_from_coords(coords: tuple, shape: tuple) -> tuple:
    return tuple((_idx_from_coord(c, s) for (c, s) in zip(coords, shape)))

def normalized_coords_from_coords(coords: tuple, shape:tuple) -> tuple:
    """Return the normalized coordinates in [0, 1] for coordinates in
    [0, shape]. Note that a coordinate (normalized or otherwise) centered on
    pixel 0 should be slightly larger than 0."""
    return tuple([c/s for (c,s) in zip(coords, shape)])

def get_centered_subarray(
    A: np.ndarray,
    center: Tuple[int],
    radius: Tuple[int]
    ) -> np.ndarray:
    """Return a subarray centered at the given coordinate with shape
    2 * radius + 1 in each dimension.  Coordinates should be in [0, shape)."""

    padding = tuple((x, x) for x in radius)
    big_A = np.pad(A, padding)
    big_center = np.array(center) + np.array(radius)
    idx = tuple(
        slice(big_center[i]-radius[i], big_center[i]+1+radius[i])
        for i in range(len(center))
    )

    return big_A[idx]

def subpixel_max_1D(A0, A1, A2) -> float:
    """Caclulate the x-coordinate of the max for a quadratic going through the
    points (0, A0), (1, A1), and (2, A2). If A1>A0 and A1>A2, this will be a
    number in [0.5, 1.5]."""

    b = -(1.5*A0 - 2*A1 + 0.5*A2)
    a = (A0 - 2*A1 + A2)/2.0

    if a == 0:
        return 1.0
    else:
        idx = -b / (2 * a)
        if idx > 1.5 or idx < 0.5 or np.isnan(idx):
            return 1.0
        else:
            return idx

def get_nearby_max(
    A: np.ndarray,
    center: Tuple[float],
    radius: Tuple[float],
    blur_sigma: Tuple[float],
    ) -> Tuple[float]:
    """Find a nearby local maximum given an array and center, both using a
    normalized (0, 1) coordinate system. It uses 1D quadratic fitting to find
    local maxima with subpixel resolution."""

    N = len(A.shape)

    center_idx = idx_from_coords(center, A.shape)
    radius_idx = idx_from_coords(radius, A.shape)
    blur_sigma_idx = idx_from_coords(blur_sigma, A.shape)
    blur_sigma_idx = tuple([max(1, sigma) for sigma in blur_sigma_idx])

    A = blur(A, blur_sigma_idx)

    local_vol = get_centered_subarray(A, center_idx, radius_idx)

    max_indices = np.where(local_vol==np.max(local_vol))
    local_vol_max_ind = [int(np.rint(np.mean(i))) for i in max_indices]

    max_ind = (np.array(center_idx)
        - np.array(radius_idx)
        + np.array(local_vol_max_ind))

    refinement_vol = get_centered_subarray(A, max_ind, (1,)*N)
    offset = np.array((0.0,)*N)
    for i in range(N):
        idx = (1,)*i + (Ellipsis,) + (1,)*(N-i-1)
        vals = refinement_vol[idx]
        offset[i] = subpixel_max_1D(*vals) - 1

    max_ind_subpixel = max_ind + offset

    return coords_from_idx(max_ind_subpixel, A.shape)





from logging import Logger
from pathlib import Path
from typing import List

from ..data import (AnnotationTable, WorldlineTable,
                    cleanup_worldlines, color_worldlines)


def renumber_worldlines(dataset: Path, annotations: AnnotationTable,
                        worldlines: WorldlineTable, window_state, arg: str,
                        logger: Logger) -> List[dict]:
    """
    Remove unused worldlines and renumber them consecutively starting from 1.
    """

    (annotations, worldlines) = cleanup_worldlines(annotations, worldlines)

    return [{
        "type": "annotations/get_annotations"
    }, {
        "type": "worldlines/get_worldlines"
    }]


def randomly_color_worldlines(dataset: Path, annotations: AnnotationTable,
                              worldlines: WorldlineTable, window_state,
                              arg: str, logger: Logger) -> List[dict]:
    """
    Randomly color all worldlines.
    """

    color_worldlines(worldlines)

    return [
        {
            "type": "worldlines/get_worldlines"
        },
    ]



from ..utils.utils import *


class Container:
    """Zephir variable container.

    Store and handle variables and properties that are shared across Zephir methods.
    """

    def __init__(self, dataset, **kwargs):

        self.dataset = Path(dataset)
        self.metadata = get_metadata(dataset)

        self.props = {
            'dataset': self.dataset,
            'img_shape': (self.metadata['shape_z'],
                          self.metadata['shape_y'],
                          self.metadata['shape_x']),
            'shape_t': self.metadata['shape_t'],
        }

        self.update(kwargs)

    def get(self, key: str):
        """Fetch variable with given name.

        :param key: name of variable to fetch.
        :return: specified variable
        """
        if key in self.props:
            return self.props[key]
        else:
            print(f'*** ERROR: variable:{key} not found in container!')
            return None

    def update(self, props: dict):
        """Update value for existing variable or add as new variable to container.props.

        :param props: dictionary of names and variables to store.
        """
        for key, value in props.items():
            self.props[key] = value
        update_checkpoint(self.dataset, {'container': self})
        return




"""
track_frame.py: track keypoints in a single frame.

Annotator GUI macro for ZephIR to track keypoints in a single frame (and its
child frames, if specified). Necessary components are loaded from checkpoint.pt
to expedite the process.
"""

import shutil
from docopt import docopt

from . import *
from ..models.container import Container
from ..utils.utils import *
from ..__version__ import __version__


def track_frame(
    dataset,
    annotation,
    t_idx,
    restrict_update=True,
    recompile_model=False,
    update_children=False):

    if len(load_checkpoint(dataset, verbose=True)) <= 1:

        if (dataset / 'args.json').is_file():
            with open(str(dataset / 'args.json')) as json_file:
                args = json.load(json_file)
        else:
            print('Creating checkpoint with default args...')
            from .. import main as zeir
            args = docopt(zeir.__doc__, argv=[f'--dataset={dataset}'])

        update_checkpoint(dataset, {'__version__': __version__, 'args': args})

        if torch.cuda.is_available():
            print('\n*** GPU available!')
            dev = 'cuda'
        # elif torch.backends.mps.is_available():
        #     dev = 'mps'
        else:
            dev = 'cpu'
        print(f'\nUsing device: {dev}\n\n')

        container = Container(
            dataset=dataset,
            allow_rotation=args['--allow_rotation'] in ['True', 'Y', 'y'],
            channel=int(args['--channel']) if args['--channel'] else None,
            dev=dev,
            exclude_self=args['--exclude_self'] in ['True', 'Y', 'y'],
            exclusive_prov=(bytes(args['--exclusive_prov'], 'utf-8')
                            if args['--exclusive_prov'] else None),
            gamma=float(args['--gamma']),
            include_all=args['--include_all'] in ['True', 'Y', 'y'],
            lr_coef=float(args['--lr_coef']),
            n_frame=int(args['--n_frame']),
            z_compensator=float(args['--z_compensator']),
        )
        recompile_model = True

    if not (dataset / 'backup').is_dir():
        Path.mkdir(dataset / 'backup')
    now = datetime.datetime.now()
    now_ = now.strftime("%m_%d_%Y_%H_%M_%S")
    shutil.copy(dataset / 'checkpoint.pt',
                dataset / 'backup' / f'checkpoint_{now_}.pt')

    print('\nFetching current results...\n')
    args = get_checkpoint(dataset, 'args')
    container = get_checkpoint(dataset, 'container')

    container.update({'exclude_self': False})
    container, _results = build_annotations(
        container=container,
        annotation=annotation,
        t_ref=None,
        wlid_ref=eval(args['--wlid_ref']) if args['--wlid_ref'] else None,
        n_ref=int(args['--n_ref']) if args['--n_ref'] else None,
    )
    container.update({'exclude_self': True})
    container, _ = build_annotations(
        container=container,
        annotation=annotation,
        t_ref=None,
        wlid_ref=eval(args['--wlid_ref']) if args['--wlid_ref'] else None,
        n_ref=int(args['--n_ref']) if args['--n_ref'] else None,
    )

    if recompile_model:
        container, zephir, zephod = build_models(
            container=container,
            dimmer_ratio=float(args['--dimmer_ratio']),
            grid_shape=(5, 2 * (int(args['--grid_shape']) // 2) + 1,
                        2 * (int(args['--grid_shape']) // 2) + 1),
            fovea_sigma=(1, float(args['--fovea_sigma']),
                         float(args['--fovea_sigma'])),
            n_chunks=int(args['--n_chunks']),
        )
        container = build_springs(
            container=container,
            load_nn=args['--load_nn'] in ['True', 'Y', 'y'],
            nn_max=int(args['--nn_max']),
        )
        container = build_tree(
            container=container,
            sort_mode=str(args['--sort_mode']),
            t_ignore=eval(args['--t_ignore']) if args['--t_ignore'] else None,
            t_track=eval(args['--t_track']) if args['--t_track'] else None,
        )
    else:
        zephir = get_checkpoint(dataset, 'zephir')
        zephod = get_checkpoint(dataset, 'zephod')

    t_list = [t_idx]
    if update_children:
        p_list = container.get('p_list')
        parents, p = [t_idx], 0
        while p < len(parents):
            children = np.where(np.array(p_list) == parents[p])[0]
            for c in children:
                t_list.append(c)
                parents.append(c)
            p += 1
    t_list = np.array(t_list, dtype=int)

    container, results = track_all(
        container=container,
        results=_results,
        zephir=zephir,
        zephod=zephod,
        clip_grad=float(args['--clip_grad']),
        lambda_t=float(args['--lambda_t']),
        lambda_d=float(args['--lambda_d']),
        lambda_n=float(args['--lambda_n']),
        lambda_n_mode=args['--lambda_n_mode'],
        lr_ceiling=float(args['--lr_ceiling']),
        lr_floor=float(args['--lr_floor']),
        motion_predict=args['--motion_predict'] in ['True', 'Y', 'y'],
        n_epoch=int(args['--n_epoch']),
        n_epoch_d=(int(args['--n_epoch_d'])
                   if float(args['--lambda_d']) > 0 else 0),
        restrict_update=restrict_update,
        _t_list=t_list,
    )

    return (
        list(t_list),
        results[t_list, ...],
        list(container.get('worldline_id')),
        container.get('provenance')[t_list, ...]
    )



import colorsys
import cv2
import matplotlib.pyplot as plt
import torch
import torch.nn.functional as F

from .io import *


# core utilities for ZephIR
def blur3d(input_tensor, kernel_size, sigmas, dev=torch.device('cpu')):
    return F.conv3d(
        input_tensor.view(-1, *input_tensor.shape[2:]),
        get_gaussian_kernel(
            kernel_size=kernel_size,
            sigmas=sigmas,
            channels=input_tensor.size(2),
            dev=dev
        ),
        groups=input_tensor.size(2),
        padding=tuple([int((size - 1) / 2) for size in kernel_size])
    ).view(input_tensor.shape)


def get_gaussian_kernel(kernel_size=(1, 3, 3), sigmas=(1, 2, 2), channels=2, dev=torch.device('cpu')):
    kernel = torch.ones(kernel_size, device=dev)
    meshgrids = torch.meshgrid(
        [torch.arange(size, device=dev) for size in kernel_size],
        indexing='ij'
    )
    for size, std, mgrid in zip(kernel_size, sigmas, meshgrids):
        mean = (size - 1) / 2
        kernel *= (1 / (std * np.sqrt(2 * np.pi))
                   * torch.exp(-torch.pow((mgrid - mean) / std, 2) / 2))

    # kernel must sum to one
    kernel = kernel / torch.sum(kernel)

    # repeat over channels
    if channels > 0:
        kernel = kernel.view(1, 1, *kernel.size())
        kernel = kernel.repeat(channels, 1, *[1] * len(kernel_size))

    return kernel


def to_tensor(data, n_dim=None, grad=False, dtype=None, dev=torch.device('cpu')):
    if isinstance(data, np.ndarray) or isinstance(data, list):
        if n_dim is not None:
            while len(data.shape) < n_dim:
                data = np.array(data)[np.newaxis, ...]
        if dtype is not None:
            return torch.tensor(data, requires_grad=grad, device=dev, dtype=dtype)
        return torch.tensor(data, requires_grad=grad, device=dev, dtype=torch.float32)
    else:
        return data


def to_numpy(tensor):
    if isinstance(tensor, torch.Tensor):
        return tensor.clone().cpu().detach().numpy()
    else:
        return tensor


def get_pixel(coords, img_shape):
    if isinstance(coords, torch.Tensor):
        return idx_from_coords(
            tuple((to_numpy(coords) + 1) / 2),
            tuple(np.array(img_shape)[::-1])
        )
    else:
        return idx_from_coords(
            tuple((coords + 1) / 2),
            tuple(np.array(img_shape)[::-1])
        )


def pick_color(x):
    return 255 * np.array(
        colorsys.hls_to_rgb(
            ((x * 157) % 360) / 360,
            (65 + (x * 7) % 35) / 100,
            (47 + (x * 7) % 53) / 100
        )
    )


def save_as_bgr(img, scale=(2, 1, 1), c=None, annotation=None, path=None, s=2):
    if np.max(img) <= 1.0:
        img *= 255

    composite = []
    for channel in range(img.shape[0]):
        if np.max(img[channel, ...]) > 1e-8:
            composite.append(
                mip_threeview(
                    np.uint8(img[channel, ...]),
                    scale
                )
            )

    if len(composite) == 3:
        r = auto_lut(composite[0], newtype=np.uint8)
        g = auto_lut(composite[1], newtype=np.uint8)
        b = auto_lut(composite[2], newtype=np.uint8)
        frame = np.dstack([r, g, b])
    elif len(composite) == 2:
        frame = compare_red_green(composite[0], composite[1])
    elif len(composite) == 1:
        if c == 0 or c is None:
            r = auto_lut(composite[0], newtype=np.uint8)
            g = np.zeros_like(r)
            b = np.zeros_like(r)
        elif c == 1:
            g = auto_lut(composite[0], newtype=np.uint8)
            r = np.zeros_like(g)
            b = np.zeros_like(g)
        elif c >= 2:
            b = auto_lut(composite[0], newtype=np.uint8)
            r = np.zeros_like(b)
            g = np.zeros_like(b)
        else:
            r = auto_lut(composite[0], newtype=np.uint8)
            g = np.zeros_like(r)
            b = np.zeros_like(r)
        frame = np.dstack([r, g, b])
    else:
        r = mip_threeview(np.uint8(img[0, ...]), scale)
        g = np.zeros_like(r)
        b = np.zeros_like(r)
        frame = np.dstack([r, g, b])

    if annotation is not None:
        for n in range(annotation.shape[0]):
            scaled_shape = img.shape[1:] * np.array(scale)
            x, y, z = get_pixel(annotation[n, ...], scaled_shape)
            color = pick_color(n)
            frame = cv2.circle(frame, (int(x), int(y)), s, color=color, thickness=-1)
            frame = cv2.circle(frame, (int(x), int(z + scaled_shape[1])), s, color=color, thickness=-1)
            frame = cv2.circle(frame, (int(z + scaled_shape[2]), int(y)), s, color=color, thickness=-1)

    if path is not None:
        cv2.imwrite(str(path), frame[:, :, ::-1])

    return frame


def live_evolve_neuron(neurons, msg):
    from IPython import display
    _ = display.clear_output(wait=True)
    _ = plt.cla()
    _ = plt.clf()
    f, axes = plt.subplots(1, len(neurons))
    for n, descriptor in enumerate(neurons):
        if descriptor.shape[0] == 2:
            descriptor = np.append(descriptor, np.zeros((1, *descriptor.shape[1:])), axis=0)
        elif descriptor.shape[0] == 1:
            descriptor = np.append(descriptor, np.zeros((2, *descriptor.shape[1:])), axis=0)
        _ = axes[n].imshow(np.max(np.transpose(descriptor, (1, 2, 3, 0)), axis=0) / np.max(descriptor))
    print(msg)
    _ = display.display(plt.gcf())


def plot_with_indicator_v(dataset, indicators, x_list=None, title=''):
    color_list = ['r', 'g', 'b', 'c', 'm', 'y']
    for i in range(len(indicators)):
        pt, start, end = indicators[i]
        color = color_list[i % len(color_list)]
        plt.vlines(pt, start, end, color=color)
    for j in range(len(dataset)):
        if x_list is None:
            plt.plot(dataset[j])
        else:
            plt.plot(x_list[j], dataset[j])
    plt.title(title)
    plt.show()


# ported over from vlab.images.transform
def _idx_from_coords(coords: float, shape: int) -> int:
    return max(round(coords*shape - 1E-6), 0)


def idx_from_coords(coords: tuple, shape: tuple) -> tuple:
    return tuple((_idx_from_coords(c, s) for (c, s) in zip(coords, shape)))


def _coords_from_idx(idx: int, shape: int) -> float:
    return (idx + 0.5) / shape


def coords_from_idx(idx: tuple, shape: tuple) -> tuple:
    return tuple((_coords_from_idx(i, s) for (i, s) in zip(idx, shape)))


def apply_lut(x: np.ndarray, lo: float, hi: float, newtype=None) -> np.ndarray:
    """Clip x to the range [lo, hi], then rescale to fill the range of
    newtype."""

    if newtype is None:
        newtype = x.dtype

    y_float = (x-lo)/(hi-lo)
    y_clipped = np.clip(y_float, 0, 1)

    if np.issubdtype(newtype, np.integer):
        maxval = np.iinfo(newtype).max
    else:
        maxval = 1.0

    return (maxval*y_clipped).astype(newtype)


def auto_lut(x: np.ndarray, quantiles=(0.5,0.99), newtype=None) -> np.ndarray:
    """Linearly map the specified quantiles of x to the range of newtype."""

    lo = np.quantile(x, quantiles[0])
    hi = np.quantile(x, quantiles[1])

    return apply_lut(x, lo, hi, newtype=newtype)


def mip_threeview(vol: np.ndarray, scale=(4,1,1)) -> np.ndarray:
    """Combine 3 maximum intensity projections of a volume into a single
    2D array."""

    s = vol.shape[:3] * np.array(scale)
    output_shape = (s[1] + s[0],
                    s[2] + s[0])

    if vol.ndim == 4:
        output_shape = (*output_shape, 3)

    vol = np.repeat(vol, scale[0], axis=0)
    vol = np.repeat(vol, scale[1], axis=1)
    vol = np.repeat(vol, scale[2], axis=2)

    x = np.transpose(np.max(vol, axis=2), (1, 0, *(range(2, np.ndim(vol)-1))))
    y = np.max(vol, axis=1)
    z = np.max(vol, axis=0)

    output = np.zeros(output_shape, dtype=vol.dtype)

    output[:s[1], :s[2]] = z
    output[:s[1], s[2]:] = x
    output[s[1]:, :s[2]] = y

    return output


def compare_red_green(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """Autoscale 2D arrays x and y, then stack them with a blank blue channel
    to form a 3D array with shape (3, size_Y, size_X) and type uint8."""

    r = auto_lut(x, newtype=np.uint8)
    g = auto_lut(y, newtype=np.uint8)
    b = np.zeros_like(r)

    return np.dstack([r, g, b])


def fill_nans(data):
    temp = data.copy()
    nans_idx = np.where(np.isnan(temp))[0]
    for i in nans_idx:
        j = i - 1
        while data[j] == np.nan:
            j += -1
        temp[i] = temp[j]
    return temp


# ported from vlab.images.order_times
def get_undiscounted_scores_for_tree(p_dist, parents, shape_t):
    scores = np.zeros(shape_t)
    for i in range(shape_t):
        if parents[i] < 0:
            continue
        scores[i] = p_dist[i, int(parents[i])]
    return scores


def get_depth(parents, node, node_depth=0):
    """Get depth of a given node given a tree formated as a list of parents,
    with the root pointing to -1."""

    parent = parents[node]
    if parent == -1:
        return node_depth
    else:
        return get_depth(parents, parent, node_depth + 1)


# utilities for Zephod
def gaussian(sigma: np.ndarray, shape=None, dtype=np.float32, norm="max"
             ) -> np.ndarray:
    """Make a 3D gaussian array density with the specified shape. norm can be
    either 'max' (largest value is set to 1.0) or 'area' (sum of values is
    1.0)."""

    sigma = np.array(sigma)
    if shape is None:
        shape = 2*sigma + 1
    shape = np.array(shape).astype(int)
    bounds = ((shape - 1)/2).astype(int)

    z = np.linspace(-bounds[0], bounds[0], shape[0])
    y = np.linspace(-bounds[1], bounds[1], shape[1])
    x = np.linspace(-bounds[2], bounds[2], shape[2])
    Z, Y, X = np.meshgrid(z, y, x, indexing='ij')
    g = np.exp(-(X**2)/(2.0*sigma[2]) - (Y**2)/(2.0*sigma[1]) - (Z**2)/(2.0*sigma[0]))

    if norm == "max":
        return (g / np.max(g)).astype(dtype)
    elif norm == "area":
        return (g / np.sum(g)).astype(dtype)
    else:
        raise ValueError("norm must be one of 'max' or 'area'")


def get_times(dataset_path: Path) -> np.ndarray:
    """Return time indices in a dataframe.
    This should return a 1-D numpy array containing integer indices for get_slice().
    """
    h5_filename = dataset_path / "data.h5"
    f = h5py.File(h5_filename, 'r')
    return np.arange(len(f["times"][:]))



import torch.nn as nn
import torch.nn.functional as F

from inspect import getmembers, isfunction

from . import channels
from ..utils import utils
from ..utils.utils import *


class ZephOD(nn.Module):
    def __init__(self,
                 n_channels_in=6,
                 n_channels_out=1,
                 init_nodes=16,
                 kernel=(1, 3, 3),
                 padding=1,
                 pool_kernel=(2, 2, 2)):

        super(ZephOD, self).__init__()

        self.n_channels_in = n_channels_in
        self.n_channels_out = n_channels_out
        self.init_nodes = init_nodes
        self.kernel = kernel
        self.padding = padding
        self.pool_kernel = pool_kernel

        self.alpha = nn.Parameter(
            torch.ones(n_channels_in) * 0.5,
            requires_grad=True
        )
        self.gamma = nn.Parameter(
            torch.ones(n_channels_in),
            requires_grad=False
        )
        self.conv1 = conv(
            self.n_channels_in,
            self.init_nodes,
            nn.ReLU(),
            self.kernel,
            self.padding
        )
        self.conv2 = conv(
            self.init_nodes,
            self.init_nodes * 2,
            nn.ReLU(),
            self.kernel,
            self.padding
        )
        self.output_layer = conv(
            self.init_nodes * 2,
            self.n_channels_out,
            nn.Sigmoid(),
            self.kernel,
            self.padding,
        )

    def preprocess(self, vol):
        with torch.no_grad():
            n_channel = vol.shape[0]
            for c in range(n_channel):
                vol = np.append(
                    vol, np.stack(
                        [f[1](vol[c]) for f in getmembers(channels, isfunction)
                         if f not in getmembers(utils, isfunction)],
                        axis=0
                    ), axis=0
                )
            for c in range(vol.shape[0]):
                vol[c] = vol[c] / (np.max(vol[c]) + 1E-8)
        return to_tensor(vol, n_dim=5, dev=self.alpha.device)

    def forward(self, input_tensor):
        if isinstance(input_tensor, np.ndarray):
            input_tensor = self.preprocess(input_tensor)
        gamma = torch.pow(input_tensor, self.gamma.view(1, -1, 1, 1, 1))
        alpha = self.alpha.view(1, -1, 1, 1, 1) * gamma
        conv1 = self.conv1(alpha)
        conv2 = self.conv2(F.max_pool3d(conv1, self.pool_kernel))
        output_tensor = self.output_layer(conv2)
        return F.interpolate(output_tensor, input_tensor.shape[2:], mode='trilinear')


def conv(n_channels_in,
         n_channels_out,
         activation,
         kernel=(1, 3, 3),
         padding=1):

    return nn.Sequential(
        nn.Conv3d(n_channels_in,
                  n_channels_out,
                  kernel,
                  padding=padding,
                  padding_mode='zeros'),
        nn.BatchNorm3d(n_channels_out),
        activation
    )



from collections import OrderedDict
import copy
import torch.nn as nn
import torch.optim as optim
from itertools import product
from scipy.interpolate import griddata

from ..models.losses import *
from ..utils.utils import *


def track_all(
    container,
    results,
    zephir,
    zephod,
    clip_grad,
    lambda_t,
    lambda_d,
    lambda_n,
    lambda_n_mode,
    lr_ceiling,
    lr_floor,
    motion_predict,
    n_epoch,
    n_epoch_d,
    kernel_size=(3, 9, 9),
    lambda_n_decay=1.0,
    lr_step_size=10,
    lr_gamma=0.5,
    nb_delta=(2, 1),
    nb_epoch=5,
    restrict_update=False,
    sigmas=(1, 4, 4),
    _t_list=None,):

    # pull variables from container
    dataset = container.get('dataset')
    allow_rotation = container.get('allow_rotation')
    annot = container.get('annot')
    channel = container.get('channel')
    covar = container.get('covar')
    dev = container.get('dev')
    gamma = container.get('gamma')
    lr_coef = container.get('lr_coef')
    n_frame = container.get('n_frame')
    neighbors = container.get('neighbors')
    partial_annot = container.get('partial_annot')
    p_list = container.get('p_list')
    r_list = container.get('r_list')
    s_list = container.get('s_list')
    shape_t = container.get('shape_t')
    t_annot = container.get('t_annot')
    t_list = container.get('t_list')
    z_compensator = container.get('z_compensator')

    if _t_list is None:
        _t_list = t_list.copy()
        update_checkpoint(dataset, {
            'results': results,
            '_t_list': _t_list,
        }, verbose=False)
    else:
        t_list = _t_list.copy()

    zephir.to(dev)
    tpbar = tqdm(t_list, desc='Analyzing movie', unit='frames')
    for t in tpbar:

        # get frame branch information
        parent = p_list[t]
        root = r_list[t]
        if parent < 0:
            # root frame, does not need registration
            results[t] = annot[root]
            continue

        # calculate initial learning rate based on parent-child distance
        distance = s_list[t]
        if lr_coef > 0 and distance > 0:
            lr_init = min(max(lr_coef * distance, lr_floor), lr_ceiling)
        else:
            lr_init = lr_floor
        tqdm.write(f'\nFrame #{t}\t\tParent #{parent}'
                   f'\t\tReference #{t_annot[root]}'
                   f'\t\tDistance to parent: d={distance:.4f}')

        if n_frame > 1:
            t_patch = np.arange(max(t-n_frame//2, 0), min(t+n_frame//2+1, shape_t))
            t_idx = np.where(t_patch == t)[0][0]
        else:
            t_patch = np.array([t])
            t_idx = 0

        with torch.no_grad():

            # compiling Zephod results for feature detection loss, L_D
            input_tensor_d = []
            if lambda_d > 0 and zephod is not None:
                for _t in t_patch:
                    data = get_data(dataset, _t, g=gamma, c=channel)
                    _pred = zephod(data)
                    _pred = (torch.max(_pred) - _pred) / torch.max(_pred)
                    input_tensor_d.append(torch.where(_pred < 0.5, 0., 0.5))
                input_tensor_d = torch.stack(input_tensor_d, dim=0)
                input_tensor_d = blur3d(input_tensor_d, (1, 5, 5), (1, 2, 2), dev=dev)
                input_tensor_d.requires_grad = False

            # compiling target descriptors from reference frame for image registration, L_R
            data = get_data(dataset, t_annot[root], g=gamma, c=channel)
            vol = to_tensor(data, n_dim=5, dev=dev)
            zephir.theta.zero_()
            zephir.rho.zero_()
            zephir.rho[:len(t_patch)].add_(
                to_tensor(annot[root], dev=dev).expand(len(t_patch), -1, -1)
            )
            target_tensor = zephir(vol.expand(len(t_patch), -1, -1, -1, -1, -1))

            # loading input volumes
            input_tensor = []
            for _t in t_patch:
                data = get_data(dataset, _t, g=gamma, c=channel)
                vol = to_tensor(data, n_dim=5, grad=False, dev=dev)
                input_tensor.append(vol)
            input_tensor = torch.stack(input_tensor, dim=0)

            # initializing model parameters at parent coordinates
            xyz_parent = to_tensor(results[parent, ...], dev=dev)
            zephir.theta.zero_()
            zephir.rho.zero_()
            zephir.rho[:len(t_patch)].add_(
                xyz_parent.expand(len(t_patch), -1, -1)
            )

            # compiling spring connections and stiffnesses for spatial regularization, L_N
            ind = neighbors[root][:, :, 0]
            d_ref = to_tensor(neighbors[root][:, :, 1:], dev=dev)
            if covar is not None:
                nn_covar = np.empty_like(ind)
                for k in range(ind.shape[1]):
                    nn_covar[:, k] = covar[range(ind.shape[0]), ind[:, k].astype(int)]/20
                k_ij = to_tensor(
                    nn_covar /
                    np.clip(np.max(nn_covar, axis=-1)[:, None], 1, None),
                    grad=False, dev=dev
                )
                k_ij = lambda_n * torch.relu(k_ij)
            elif torch.max(d_ref) < 0:
                k_ij = torch.zeros_like(
                    to_tensor(ind), requires_grad=False, device=dev
                )
            else:
                k_ij = lambda_n * torch.ones_like(
                    to_tensor(ind), requires_grad=False, device=dev
                )

            # loading partial annotations if available
            p_idx, pins, params = [], [], []
            subset = _rho = None
            for i, _t in enumerate(t_patch):
                if _t in partial_annot.keys():
                    tqdm.write(f'*** Partial annotations available for frame {_t}')
                    _pins, _params = partial_annot[_t]

                    if motion_predict:
                        # using partial annotations to interpolate a flow field
                        # and calculate new initial parameters for tracking
                        corners = np.stack(list(product((-1, 1), repeat=3)), axis=0)
                        p = np.append(results[parent, _pins, :], corners, axis=0)
                        v = np.append(_params - results[parent, _pins, :], np.zeros((8, 3)), axis=0)
                        gd = griddata(p, v, results[parent, ...], method='linear', fill_value=0)
                        zephir.rho[i].add_(to_tensor(gd, dev=dev))

                    p_idx.append(i)
                    pins.append(to_tensor(_pins, dtype=torch.int64, dev=dev))
                    params.append(to_tensor(_params, dev=dev))

            if restrict_update and t in partial_annot.keys():
                # restricting tracking to only keypoints neighboring partial annotations
                _pins, _params = partial_annot[t]
                subset = np.unique(np.append(_pins, ind[_pins, :])).astype(int)
                tqdm.write(f'*** RESTRICT_UPDATE ACTIVE. '
                           f'Deepcopying model with restricted parameters '
                           f'for {len(subset)} keypoints...')
                zephir = copy.deepcopy(zephir)
                zephir.shape_n = len(subset)
                zephir.grid = nn.Parameter(zephir.grid[:, subset], requires_grad=False)
                zephir.rho = nn.Parameter(zephir.rho[:, subset], requires_grad=True)
                zephir.theta = nn.Parameter(zephir.theta[:, subset], requires_grad=allow_rotation)
                target_tensor = target_tensor[:, subset]
                _rho = to_tensor(results[t_patch], dev=dev, grad=False)
                _rho[:, subset] = zephir.rho.clone()
                _rho.requires_grad = False

            # resetting optimizer state
            optimizer = optim.SGD(zephir.parameters(), lr=lr_init)
            scheduler = optim.lr_scheduler.StepLR(optimizer, lr_step_size, lr_gamma)
            _kernel_size = kernel_size
            _sigmas = sigmas

        # training loop here
        zephir.train()
        pbar = tqdm(range(n_epoch + n_epoch_d), desc='Tracking', unit='epochs')
        for epoch in pbar:

            if (epoch + 1) % nb_epoch == 0:
                with torch.no_grad():
                    # updating spring stiffness
                    if lambda_n_decay < 1.0:
                        k_ij *= lambda_n_decay
                    # updating kernels for dynamic Gaussian blurring
                    _kernel_size = (3, _kernel_size[1] - nb_delta[0], _kernel_size[2] - nb_delta[0])
                    _sigmas = (1, _sigmas[1] - nb_delta[1], _sigmas[1] - nb_delta[1])
                    if _kernel_size[1] <= 1 or _sigmas[1] <= 1:
                        # resets kernels to be inactive when it reaches a threshold
                        _kernel_size = (1, 1, 1)
                        _sigmas = (1, 1, 1)

            optimizer.zero_grad()

            # using automatic matrix precision casting to dynamically switch
            # between float16 and float32 to minimize memory consumption
            with torch.autocast(device_type=dev):

                # sampling child descriptors at model parameters from child frames
                pred = zephir(input_tensor)

                # dynamic Gaussian blurring
                blur_target = blur3d(target_tensor, _kernel_size, _sigmas, dev=dev)
                blur_pred = blur3d(pred, _kernel_size, _sigmas, dev=dev)

                if epoch > n_epoch and lambda_d > 0 and input_tensor_d is not None:
                    # feature detection loss, L_D
                    loss = torch.mean(reg_d(
                        lambda_d, input_tensor_d,
                        zephir.rho, zephir.n_chunks
                    ))
                else:
                    # image registration loss, L_R
                    loss = torch.mean(corr_loss(blur_pred, blur_target))

                    reg = 0
                    if torch.max(k_ij) > 0:
                        # spatial regularization, L_N
                        reg += torch.mean(
                            reg_n(k_ij, zephir.rho, _rho, ind, d_ref,
                                  subset, lambda_n_mode)
                        )
                    if lambda_t > 0:
                        # temporal smoothing, L_T
                        reg += torch.mean(reg_t(lambda_t, blur_pred, (1, 5, 5), None))
                    loss += reg

            # backpropagation call
            loss.backward()
            if clip_grad > 0.:
                nn.utils.clip_grad_value_(zephir.parameters(), clip_grad)
            if z_compensator > 0.:
                # due to anisotropy in depth-axis commonly found in 3D datasets,
                # gradient descent may not properly update parameters in that axis
                # so we multiply calculated gradients in that axis by (1+z_compensator)
                # to compensate
                with torch.no_grad():
                    zephir.rho[:, :, -1] += (
                        - z_compensator
                        * optimizer.param_groups[0]['lr']
                        * zephir.rho.grad[:, :, -1]
                    )

            # gradient descent step, update model parameters
            optimizer.step()
            scheduler.step()

            with torch.no_grad():
                current_lr = optimizer.param_groups[0]['lr']
                pbar.set_postfix(OrderedDict(
                    Loss=f'{loss.item():.4f}',
                    Loss_R=f'{(loss - reg).item():.4f}',
                    LR=f'{current_lr:.4f}'
                ))

                # resetting model parameters for partial annotations
                if len(p_idx) > 0:
                    for _i, i in enumerate(p_idx):
                        _pins, _params = pins[_i], params[_i]
                        if subset is not None:
                            _pins = [np.where(subset == w)[0][-1]
                                     for w in to_numpy(_pins) if w in subset]
                            _pins = to_tensor(_pins, dtype=torch.int64, dev=dev)
                        zephir.rho[i].index_fill_(0, _pins, 0)
                        zephir.rho[i].index_add_(0, _pins, _params)

                if subset is not None:
                    _rho[:, subset] = zephir.rho.clone()
                    _rho.requires_grad = False
                
                # resetting model parameters for rotation
                if not allow_rotation:
                    zephir.theta.zero_()

        # updating array with tracking results for the child frame
        with torch.no_grad():
            if subset is not None:
                _rho[t_idx, subset, :] = zephir.rho[t_idx]
                results[t] = to_numpy(_rho[t_idx])
            else:
                results[t] = to_numpy(zephir.rho)[t_idx]

            _t_list = np.setdiff1d(_t_list, np.array([t]), assume_unique=True)

            # push results to checkpoint
            update_checkpoint(dataset, {
                'results': results,
                '_t_list': _t_list,
            }, verbose=False)

    return container, results



from logging import Logger
from pathlib import Path
from typing import List

from ..data import AnnotationTable, WorldlineTable
from ...methods.overwrite_checkpoint import overwrite_checkpoint

from ._utilities import default_args


@default_args("None, None")
def overwrite_zeir_checkpoint(
    dataset: Path,
    annotations: AnnotationTable,
    worldlines: WorldlineTable,
    window_state,
    arg: str,
    logger: Logger
) -> List[dict]:
	"""
	arg: key, value
		key: Name of item in checkpoint to overwrite.
		value: New item to write into checkpoint.
	ex: "kn_max", "10": Updates max number of neighbor spring connections to 10.
		"clip_grad", "0.1": Updates gradient clipping ceiling to 0.1.
	"""

	arg_list = arg.replace(" ", "").split(",")

	key = str(arg_list[0])
	value = eval(arg_list[1])

	overwrite_checkpoint(dataset, key, value)

	return [{"type": "annotations/get_annotations"}]



# Author: vivekv2@gmail.com

from typing import Callable

def default_args(default_args: str) -> Callable:
    """This returns a decorator that attaches a default_args field to the
    function object, which is used to populate the arg textfield in the app."""

    def dec(fn: Callable) -> Callable:
        fn.default_args = default_args
        return fn

    return dec



# This provides functions that create shapes for generating synthetic 3D
# shapes.
#
# Author: vivekv2@gmail.com

from typing import Tuple

import numpy as np

def get_centered_zyx_meshgrid(shape: np.ndarray) -> Tuple[np.ndarray, ...]:
    """Return a tuple (Z,Y,X) of coordinates for a 3D array generated via
    np.meshgrid. shape should be an array of odd numbers, so the center value
    of the coordinate box can be (0,0,0)."""

    shape = np.array(shape)

    bounds = (shape - 1)/2

    z = np.linspace(-bounds[0], bounds[0], shape[0])
    y = np.linspace(-bounds[1], bounds[1], shape[1])
    x = np.linspace(-bounds[2], bounds[2], shape[2])

    Z, Y, X = np.meshgrid(z, y, x, indexing='ij')

    return (Z, Y, X)

def gaussian(sigma:np.ndarray, shape=None, dtype=np.float32, norm="max"
) -> np.ndarray:
    """Make a 3D gaussian array density with the specified shape. norm can be
    either 'max' (largest value is set to 1.0) or 'area' (sum of values is
    1.0)."""

    sigma = np.array(sigma)

    if shape is None:
        shape = 2*sigma + 1

    (Z, Y, X) = get_centered_zyx_meshgrid(shape)

    g = np.exp(-(X**2)/(2.0*sigma[2])
               -(Y**2)/(2.0*sigma[1])
               -(Z**2)/(2.0*sigma[0]))

    if norm=="max":
        g = g / np.max(g)
    elif norm=="area":
        g = g / np.sum(g)
    else:
        raise ValueError("norm must be one of 'max' or 'area'")

    return g.astype(dtype)

def ellipsoid(r:np.ndarray, shape=None, dtype=np.float32) -> np.ndarray:
    """Make an ellipse with the specified radius."""

    r = np.array(r)
    if shape is None:
        shape = 2*r + 1

    (Z, Y, X) = get_centered_zyx_meshgrid(shape)

    D = (Z/r[0])**2 + (Y/r[1])**2 + (X/r[2])**2

    return (D < 1).astype(dtype)

def place_pattern(pattern:np.ndarray, location: np.ndarray, vol: np.ndarray
) -> None:
    """Place the pattern array in vol (mutating) centered at location. For good
    centering, pattern should have odd dimension."""

    location = np.array(location)

    s = (location - (np.array(pattern.shape) - 1)/2).astype(int)
    e = (s + pattern.shape).astype(int)
    vol[s[0]:e[0], s[1]:e[1], s[2]:e[2]] = pattern

def place_patterns(patterns: Tuple[np.ndarray, ...],
        locations: Tuple[np.ndarray, ...], vol: np.ndarray) -> None:
    """Place each pattern in vol (mutating) centered at the corresponding
    locations."""

    for i in len(patterns):
        place_pattern(patterns[i], locations[i], vol)


